// common
import { RandomInt } from "../../../helpers/random";
// events
import * as EV from "../../game-events";
// trade
import { StoredResource } from "../trade/resource-storage";
import { POP_CLASSES } from './population-class';
/* == CITY POPS MODEL == */
var CityPopulation = /** @class */ (function () {
    function CityPopulation(city) {
        // population fields
        this.classes = POP_CLASSES;
        this.classPops = [];
        this.classStartingPops = [];
        // pop growth
        this.totalGrowthIncrease = 0;
        // supply & consumtion
        this.lastConsumptionSuccess = 0;
        this.lastSuppliedResIds = [];
        this.city = city;
        this.generateClassPops();
    }
    CityPopulation.prototype.generateClassPops = function () {
        for (var classNum = 0; classNum < POP_CLASSES.length; classNum++) {
            // retrieve max starting pops and pick a random amount below
            var maxPopsForClass = POP_CLASSES_MAX_START[classNum];
            var rndClassPops = RandomInt.below(maxPopsForClass);
            // save starting population
            this.classStartingPops.push(rndClassPops);
            // add to class list
            this.classPops.push(rndClassPops);
        }
        this.startingClassPops = this.classPops;
    };
    CityPopulation.prototype.getPopsOf = function (popClassLevel) {
        return this.classPops[popClassLevel];
    };
    Object.defineProperty(CityPopulation.prototype, "totalPops", {
        get: function () {
            var totalPops = 0;
            for (var _i = 0, _a = this.classPops; _i < _a.length; _i++) {
                var classPops = _a[_i];
                totalPops += classPops;
            }
            return totalPops;
        },
        enumerable: false,
        configurable: true
    });
    CityPopulation.prototype.increaseRandom = function () {
        var rndClassNum = RandomInt.below(POP_CLASSES.length);
        this.classPops[rndClassNum] += 1;
        // fire pop increased event
        this.onPopIncreased(rndClassNum);
    };
    Object.defineProperty(CityPopulation.prototype, "upgradePercent", {
        get: function () {
            var currentGrowth = this.city.growth;
            return currentGrowth / POPS_UPDATE_GROWTH * 100;
        },
        enumerable: false,
        configurable: true
    });
    /* POPULATION CONSUMPTION */
    CityPopulation.prototype.consumeResources = function () {
        // clear last consumed resource array
        this.lastSuppliedResIds = [];
        var consumed = this.getConsumedResources();
        var totalResCount = consumed.length;
        var successfulResCount = 0;
        for (var _i = 0, consumed_1 = consumed; _i < consumed_1.length; _i++) {
            var storedRes = consumed_1[_i];
            var consumedResAmount = storedRes.amount;
            var consumedResId = storedRes.resId;
            // market has resources
            if (this.market.hasEnough(consumedResId, consumedResAmount)) {
                successfulResCount++;
                this.market.remove(consumedResId, consumedResAmount);
                // add to successfully consumed respurces array
                this.lastSuppliedResIds.push(consumedResId);
            }
            else {
                // increase price anyway
                this.market.increasePrice(consumedResId, 1);
            }
        }
        var successful = successfulResCount / totalResCount;
        this.onConsumptionFinished(successful);
        return successful;
    };
    CityPopulation.prototype.resIsFullySupplied = function (resId) {
        if (this.lastSuppliedResIds.includes(resId)) {
            return true;
        }
        return false;
    };
    CityPopulation.prototype.onConsumptionFinished = function (successful) {
        this.lastConsumptionSuccess = successful;
        this.city.growth += successful;
        this.totalGrowthIncrease += successful;
    };
    CityPopulation.prototype.onPopIncreased = function (classNum) {
        var newClassPops = this.getAmount(classNum);
        globalThis.EVENTS.fireEvent(new EV.CityPopulationIncreasedEvent(this.city, classNum, newClassPops));
    };
    /* returns the amount of resources of "resId" that the popluation class "classNum" consumes in one turn */
    CityPopulation.prototype.getClassResConsumption = function (classNum, resId) {
        var popClass = this.classes[classNum];
        var consumesPer10 = popClass.getResourceConsumption(resId);
        var consumesSingle = consumesPer10 / 10;
        var consumesTotalFloat = this.getAmount(classNum) * consumesSingle;
        return consumesTotalFloat;
    };
    CityPopulation.prototype.getTotalResConsumption = function (resId) {
        var consumesTotal = 0;
        for (var _i = 0, _a = this.classes; _i < _a.length; _i++) {
            var popClass = _a[_i];
            var classConsumes = this.getClassResConsumption(popClass.level, resId);
            consumesTotal += classConsumes;
        }
        return consumesTotal;
    };
    CityPopulation.prototype.getTotalResConsumptionList = function () {
        var consumedResIds = [];
        for (var _i = 0, _a = this.classes; _i < _a.length; _i++) {
            var popClass = _a[_i];
            var classConsumed = popClass.getConsumedResIds();
            for (var _b = 0, classConsumed_1 = classConsumed; _b < classConsumed_1.length; _b++) {
                var classResId = classConsumed_1[_b];
                // add resId when not already in array (no doubles)
                var isNewResId = !consumedResIds.includes(classResId) && classResId;
                if (isNewResId) {
                    consumedResIds.push(classResId);
                }
            }
        }
        return consumedResIds;
    };
    CityPopulation.prototype.getConsumedResources = function () {
        var consumedResources = [];
        for (var _i = 0, _a = this.getTotalResConsumptionList(); _i < _a.length; _i++) {
            var consumedResId = _a[_i];
            var consumedAmount = this.getTotalResConsumption(consumedResId);
            var consumedRes = new StoredResource(consumedResId, consumedAmount);
            consumedResources.push(consumedRes);
        }
        return consumedResources;
    };
    /* returns the amount of taxes generated per turn from class pops */
    CityPopulation.prototype.getTotalTaxes = function () {
        var totalTaxes = 0;
        for (var _i = 0, _a = this.classes; _i < _a.length; _i++) {
            var popClass = _a[_i];
            var classPops = this.getPopsOf(popClass.level);
            var classTaxes = classPops * popClass.taxRate * TAXES_MODIFIER;
            totalTaxes += classTaxes;
        }
        return Math.round(totalTaxes);
    };
    Object.defineProperty(CityPopulation.prototype, "market", {
        /* POPULATION GETTERS & SETTERS */
        get: function () { return this.city.tradeNode; },
        enumerable: false,
        configurable: true
    });
    CityPopulation.prototype.getAmount = function (classNum) { return this.classPops[classNum]; };
    CityPopulation.prototype.getName = function (classNum) { return this.classes[classNum].name; };
    CityPopulation.prototype.setClassPops = function (classNum, classPops) { this.classPops[classNum] = classPops; };
    return CityPopulation;
}());
export { CityPopulation };
export var POP_CLASSES_MAX_START = [4, 6, 6, 4, 2, 2];
/* == CONSTANTS == */
export var RES_CONSUMPTION_TURNS = 7;
export var POPS_UPDATE_GROWTH = 15;
export var TAXES_MODIFIER = 1;
//# sourceMappingURL=population.js.map