// common
import * as $ from "jquery";
import { GameOptions } from './game/game-options';
import { DawnbringerPalette } from './common/colors';
// world
import { WorldMap } from './game/world/map';
import WorldMapScene from './game/scenes/world-scene';
// entity types
import { ResourceTypes } from './game/entities/trade/resource/resource-types';
import { UnitTypes } from './game/entities/unit/type/unit-types';
// entities
import { Player } from './game/entities/player/player';
// managers
import { FactionManager } from './game/managers/faction-manager';
import { EventManager } from './common/events';
import { SelectionManager } from './game/managers/selection-manager';
import { InputManager } from './game/managers/input-manager';
import { TurnManager } from './game/managers/turn-manager';
import { ConnectionManager } from "./game/managers/connection-manager";
import { AssetsLoader } from './game/managers/asset-loader';
import { ArmyManager } from './game/managers/army-manager';
import { UnitManager } from './game/managers/unit-manager';
import { CityManager } from './game/managers/city-manager';
import { TradeManager } from './game/managers/trade-manager';
import { CharacterManager } from './game/managers/character-manager';
import { UnitGearManager } from './game/managers/unit-gear-manager';
// ui
import { GuiManager } from './game/managers/gui-manager';
import { IDialog } from './game/gui/dialogs/common/base-dialog';
import { DomNodes } from './game/gui/elements/controls-dom';
// babylon
import * as BABYLON from 'babylonjs';
import "@babylonjs/loaders/glTF";
import { TradeResource } from "./game/entities/trade/resource/resource";
import { BuildingManager } from "./game/managers/building-manager";
import { FileManager } from "./game/managers/file-manager";
import { PlayerManager } from "./game/managers/player-manager";
import {ConstructableTypes} from "./game/entities/building/type/constructable";

declare global {
  var OPTIONS: GameOptions;
  var GAME_HAS_STARTED: boolean;
  var COLORS: DawnbringerPalette;
  // strings
  var CHAR_NAMES: {};
  var GEAR_TYPES: {};
  // map
  var MAP_NAME: string;
  var MAP: WorldMap;
  var WORLD: WorldMapScene;
  // scene
  var SCENE: BABYLON.Scene;
  var CAMERA: BABYLON.ArcRotateCamera;
  // types
  var RESOURCE_TYPES: ResourceTypes;
  var SPECIAL_RES: Array<TradeResource>;
  var UNIT_TYPES: UnitTypes;
  var CONSTRUCTABLE: ConstructableTypes;
  var UNIT_TYPES: UnitTypes;
  // instances
  var PLAYER: Player;
  // gui
  var GUI: GuiManager;
  var DIALOG: IDialog;
  var DOM_NODES: DomNodes;
  // assets
  var MODEL_NAMES: string[];
  var ASSETS: AssetsLoader;
  // managers
  var EVENTS: EventManager;
  var FILES: FileManager;
  var SELECTION: SelectionManager;
  var INPUT: InputManager;
  var TURNS: TurnManager;
  var PLAYER_M: PlayerManager;
  var ARMIES: ArmyManager;
  var UNITS: UnitManager;
  var GEAR: UnitGearManager;
  var CHARS: CharacterManager;
  var CITIES: CityManager;
  var TRADE: TradeManager;
  var BLD: BuildingManager;
  var CONN: ConnectionManager;
  var FACTIONS: FactionManager;
}

const logStr = "<App> ";
globalThis.GAME_HAS_STARTED = false;
globalThis.COLORS = DawnbringerPalette;
loadOptions();

window.addEventListener('DOMContentLoaded', () => {
    startEventAndFileManager();
    // default data json files
    loadDefaultData();
    // initialize resource & unit types
    initializeEntityTypes();
    createEntityManagers();
    // create UI
    createGuiDom();
    // Create the game using the 'renderCanvas'
    createScene();
    startAssetLoading();
    afterAssetsLoaded();
  });

  function loadOptions() {
    globalThis.OPTIONS = new GameOptions();
  }

function loadDefaultData() {
    _loadDefaultBldTypes();
    _loadDefaultMountTypes();
    _loadDefaultCharNames();
    _loadDefaultGearTypes();
  }

  /* == LOADING DEFAULT DATA JSON FILES == */
  function _loadDefaultCharNames() {
    const jsonPath = globalThis.FILES.DEFAULT_LANG_JSON_PATH;
    console.log(logStr + `start loading character names from ${jsonPath}`);
    $.getJSON( jsonPath)
    .done( function (data) {
      globalThis.FILES._onJsonLoaded(jsonPath, data);
      if (! globalThis.CHARS) {
        console.warn(logStr + `ERROR: cant finish char name list loading; CharacterManager not found`)
      } else {
        globalThis.CHARS.onLanguagesLoaded(data);
      }
    })
    .fail( function() { globalThis.FILES.onLoadingError(jsonPath); });
  }

  function _loadDefaultGearTypes() {
    const jsonPath = globalThis.FILES.DEFAULT_GEAR_JSON_PATH;
    console.log(logStr + `start loading gear data from ${jsonPath}`);
    $.getJSON(jsonPath)
    .done( function(data) {
      globalThis.FILES._onJsonLoaded(jsonPath, data);
      if (! globalThis.GEAR) {
        console.warn(logStr + `ERROR: cant finish gear list loading; UnitGearManager not found`)
      } else {
        globalThis.GEAR.onGearListLoaded(data);
      }
    })
    .fail( function() { globalThis.FILES.onLoadingError(jsonPath); });
  }

function _loadDefaultMountTypes() {
    const jsonPath = globalThis.FILES.DEFAULT_MOUNTS_JSON_PATH
    $.getJSON(jsonPath).
        done(data => {
            globalThis.FILES._onJsonLoaded(jsonPath, data);
            if ( !globalThis.GEAR ) {
                console.warn(logStr + `ERROR: cant finish mount list loading; UnitGearManager not found`)
            } else { globalThis.GEAR.onMountListLoaded(data); }
        })
    .fail( function() { globalThis.FILES.onLoadingError(jsonPath); });
}

function _loadDefaultBldTypes() {
    $.getJSON(globalThis.FILES.DEFAULT_BLDTYPES_JSON_PATH)
        .done(data => {
            globalThis.BLD._onJsonLoaded(data);
        });
}

  function startEventAndFileManager() {
    globalThis.EVENTS = new EventManager();
    globalThis.FILES = new FileManager();
  }

  function startAssetLoading() {
    globalThis.ASSETS = new AssetsLoader();
    globalThis.ASSETS.loadAssets();
  }

  function afterAssetsLoaded() {
    initializeInput();
    showGui();
    // Start render loop.
    globalThis.GAME_HAS_STARTED = true;
    render();
  }

  function initializeEntityTypes() {
    console.log(logStr + "initializing entity types ..");
    globalThis.RESOURCE_TYPES = new ResourceTypes();
    globalThis.SPECIAL_RES = [];
    globalThis.UNIT_TYPES = new UnitTypes();
    globalThis.CONSTRUCTABLE = new ConstructableTypes();
  }

  function createEntityManagers() {
    console.log(logStr + "creating entity managers ..");
    // before CityManager
    globalThis.SELECTION = new SelectionManager();
    globalThis.CHARS = new CharacterManager();
    globalThis.TRADE = new TradeManager();
    globalThis.BLD = new BuildingManager();
    globalThis.CITIES = new CityManager();
    // after CityManager
    globalThis.CONN = new ConnectionManager();
    globalThis.FACTIONS = new FactionManager();
    // units & gear
    globalThis.UNITS = new UnitManager();
    globalThis.GEAR = new UnitGearManager();
    globalThis.ARMIES = new ArmyManager();
    globalThis.PLAYER_M = new PlayerManager();
  }

  function createScene() {
    console.log(logStr + "creating scene ..");
    // load world
    globalThis.MAP_NAME = "world_large";
    globalThis.WORLD = new WorldMapScene('renderCanvas');
    // show the scene.
    globalThis.TURNS = new TurnManager();
    globalThis.WORLD.make();
  }

  function initializeInput() {
    globalThis.INPUT = new InputManager();
    globalThis.SCENE.onKeyboardObservable.add((kbInfo) => {
      if (kbInfo.type == BABYLON.KeyboardEventTypes.KEYUP) {
        let key: string = kbInfo.event.code.replace("Key", "");
        globalThis.INPUT.onKeyUp(key);
      }
    });
    globalThis.SCENE.onPointerDown = function (event: PointerEvent, pickResult) {
      globalThis.INPUT.onMouseClick(event, pickResult);
    }
  }

  function render() {
    globalThis.WORLD.doRender();
  }

  function createGuiDom() {
    // initialize properties
    globalThis.DIALOG = null;
    // start gui manager
    globalThis.GUI = new GuiManager();
  }

  function showGui() {
    globalThis.GUI.build();
    globalThis.GUI.showStartingMenues();
  }