// common
import { BaseEventListener } from "./events";
import { ICommandable, IBaseCommand } from "./command";
import { EntityHeader, IEntityHeader } from "./entity-header";
// world
import { WorldPos } from "../game/world/pos";
import { Loggable } from "./log";
// babylon 
import * as BABYLON from 'babylonjs';
import { WorldTools } from "../helpers/world-tools";

export interface IUniqueInstance {
    _entityType: string;
    num: number;
    uid: string;
}

/* ENTITY INTERFACES */
export interface IEntity extends IUniqueInstance {
    header: IEntityHeader;
    onTurn();
}

export interface IOnTurnEntity extends  IEntity {
    onTurn();
}

export interface IWorldEntity extends IOnTurnEntity {
    pos: WorldPos;
    mesh: BABYLON.AbstractMesh;
    distanceToPlayer: number;
}

export interface ICommandableWorldEntity extends IWorldEntity, ICommandable {

}

export interface ICommandableMovableEntity extends ICommandableWorldEntity {
    moveInDir(dirStr: string);
}

/* BASE ENTITY */ 
export class AbstractEntity extends Loggable implements IOnTurnEntity {
    _entityType: string = "abstract";
    header: IEntityHeader;
    num: number = -1;

    constructor(num: number) {
        super("Entity");
        this.num = num;
        this.header = new EntityHeader(this);
    }

    onTurn() {
        
    }

    get uid(): string { return this.header.uid(); }
}

export class AbstractEventEntity extends BaseEventListener implements IEntity {
    _entityType: string = "abstract-event";
    header: IEntityHeader;
    num: number = -1;

    constructor(num: number) {
        super("Entity");
        this.num = num;
        this.header = new EntityHeader(this);
    }

    onTurn() {
        
    }

    get uid(): string { return this.header.uid(); }
}

/* MAP ENTITY */
export class WorldEntity extends AbstractEntity implements IWorldEntity {
    pos: WorldPos;
    mesh: BABYLON.AbstractMesh;

    constructor(num: number, pos: WorldPos) {
        super(num);
        this.pos = pos;
    }

    get distanceToPlayer(): number {
        if (globalThis.PLAYER && globalThis.Player.pos) { 
            return WorldTools.getWorldPosDistance(this.pos, globalThis.PLAYER.pos); }
        return;
    }

    onTurn() {

    }
}

export class WorldEventEntity extends AbstractEventEntity implements IWorldEntity {
    pos: WorldPos;
    mesh: BABYLON.AbstractMesh;

    constructor(num: number, pos: WorldPos) {
        super(num);
        this.pos = pos;
    }

    get distanceToPlayer(): number {
        if (globalThis.PLAYER && globalThis.Player.pos) { 
            return WorldTools.getWorldPosDistance(this.pos, globalThis.PLAYER.pos); }
        return;
    }
}

/* COMMANDABLE */
export class CommandableMapEntity extends WorldEntity implements ICommandableWorldEntity {
    public commands: Array<IBaseCommand> = [];

    public onCommand(cmd: IBaseCommand) {

    }

    public addCommand(cmd: IBaseCommand) {

    }
}

/* MOVABLE ENTITY */
export class MovableEntity extends CommandableMapEntity implements ICommandableMovableEntity {
    public moveInDir(dirStr: string) {

    }
}