// common
import { IBaseEvent } from "../../../common/events";
import { WorldEventEntity } from "../../../common/entity";
import * as EV from "../../game-events";
// world
import { WorldTile } from "../../world/tile/tile";
import { TerrainTypes } from "../../world/tile/tile-terrain";
import { WorldPos, HexDirection } from "../../world/pos";
import { WorldTools } from "../../../helpers/world-tools";
// entities
import { IUniqueInstance } from "../../../common/entity";
import { City } from "../city/city";
import { Family } from "../faction/family";
import { UnitArmy } from "../unit/army/army";
// trade
import { StoredResource } from "../trade/resource/resource-storage";
import { TradeNode } from "../trade/trade-node";
// babylon
import * as BABYLON from 'babylonjs';

export interface IPlayer extends IUniqueInstance {
    num: number;
    pos: WorldPos;
    coins: number;
    localCity: City;
    army: UnitArmy;
    family: Family;
    tile: WorldTile;
    // mesh
    mesh: BABYLON.AbstractMesh;
    // methods
    showDummy(): void;
    checkCityEnterLeave(): void;
}

export class Player extends WorldEventEntity implements IPlayer {
    num: number = -1;
    pos: WorldPos = new WorldPos(100, 50);
    coins: number = STARTING_MONEY;
    mesh: BABYLON.AbstractMesh;
    army: UnitArmy;
    family: Family = null;

    constructor() {
        super(0, null);
        this.getFamily();
        this.setStartPos();
        this.showDummy();
    }

    // PLAYER START POSITION
    setStartPos() {
        const mapStartPos = globalThis.MAP.playerStartingPos;
        this.log(`setting starting pos selected from map at ${mapStartPos.asString()}`);
        this.pos = mapStartPos;
        if (this.family && this.family.customStartPos) {
            this.pos = this.family.customStartPos;
        }  
    }

    setStartPosFromHome() {
        const playerHome = this.family.homeCity;
        if ((! this.family) || (! this.family.homeCity ) ) {
            this.error(`cant get player starting pos; player family home is not set`);
            return;
        }
        this.log(`setting starting pos to ${this.family.homeCity.name} at ${this.family.homeCity.pos.asString()}`);
        this.pos = this.family.homeCity.pos;
    }

    getFamily() {
        if(! globalThis.FACTIONS ) {
            this.error(`cant get player family; FactionManager hasnt loaded yet`);
            return;
        }
        this.family = globalThis.FACTIONS.playerFamily;
    }

    // PLAYER 3D MODEL
    /* shows a sphere on player position before models are loaded */
    showDummy(): void {
        if (! globalThis.SCENE || ! globalThis.WORLD) {
            this.error(`cant create; scene or world isnt loaded yet`);
        }
        const sphereOptions = {
            diameter: 2
        };
        // create sphere dummy mesh
        this.mesh = BABYLON.MeshBuilder.CreateSphere("player-mesh", sphereOptions, globalThis.SCENE);
        // with red standard mat
        let mat: BABYLON.StandardMaterial = new BABYLON.StandardMaterial("player-mat", globalThis.SCENE);
        mat.diffuseColor = new BABYLON.Color3(1, 0, 0);
        this.mesh.material = mat;
        // at player position
        const scenePos = this.scenePos;
        scenePos.y += .5;
        this.mesh.position = scenePos;
        // inform managers
        if (globalThis.SELECTION) {
            globalThis.SELECTION.setSourceToPlayer();
        }
    }

    /* updates player model to horseman on land or small ship on water */
    updateMesh() {
        // save pos & remove old mesh
        const meshPos = this.mesh.position;
        this.mesh.dispose();
        const tile: WorldTile = globalThis.MAP.tileAtPos(this.pos);
        const isOnLand = tile.isOnLand;
        // get new mesh depending whether player is on land or water
        if (isOnLand) {
            this.mesh = globalThis.ASSETS.getAsset(ASSET_LAND).getMesh().clone("player", null);
        } else {
            this.mesh = globalThis.ASSETS.getAsset(ASSET_WATER).getMesh().clone("player", null) as BABYLON.Mesh;
        }
        this.mesh.position = meshPos;
        this.mesh.scaling = new BABYLON.Vector3(2, 2, 2);
        // set the camera target to newly created mesh
        globalThis.WORLD.renewCameraTarget(this.mesh);
        // inform managers
        if (globalThis.SELECTION) {
            globalThis.SELECTION.setSourceToPlayer();
        }
    }

    setArmy(army: UnitArmy) {
        this.army = army;
    }

    // PLAYER MOVEMENT
    moveInDir(dirStr: string) {
        // 1. save latest position in constant
        const startPos: WorldPos = this.pos;
        // 2. calculate new relative and absolute target position
        const relPos = new HexDirection(dirStr).relPos;
        const targetPos = new WorldPos(this.pos.x + relPos.x, this.pos.z + relPos.z);
        this.setPos(targetPos, dirStr, startPos);
    }

    setPos(targetPos: WorldPos, dirStr?: string, startPos?: WorldPos) {
        if (! globalThis.PLAYER_M) { return; }
        // inform managers
        globalThis.PLAYER_M.beforePlayerMovement(targetPos);
        // check possibility
        const canMove = globalThis.PLAYER_M.canMoveTo(targetPos);
        if (! canMove) { return; }
        globalThis.PLAYER_M.beforeValidPlayerMovement(targetPos);
        // 3. fire player moved event
        globalThis.EVENTS.fireEvent( new EV.PlayerMovedEvent(targetPos, dirStr, startPos));
        this.pos = targetPos;
        // inform managers
        globalThis.PLAYER_M.afterPlayerMovement();
    }

    /* appliest post movement changes: enters or leaves city? */
    checkCityEnterLeave(): void {
        /* add after movement checks here */
        if (globalThis.GUI) { globalThis.GUI.onPlayerMoved(); }
        // entered city?
        if (this.localCity) {
            // this.isInCity = cityAt;
            globalThis.EVENTS.fireEvent( new EV.CityEnteredEvent(this.localCity) );
        } else if (this.localCity) {
            // todo fix leaving
            // player left city: tell city that player has left
            // this.isInCity.onLeave(this);
            //this.isInCity = null;
        }
    }

    updatePlayerYPos() {
        this.mesh.position = this.scenePos;
        this.mesh.position.y += this.groundOffset;
    }

    /* returns height between world 0 and terrain top as number in world space */
    get groundOffset(): number {
        let heightOffset = Player.HEIGHT_OFFSET;
        if (this.tile) {
            // terrain offset: half of tile height
            heightOffset += ( TerrainTypes.heightOf(this.tile.terrainChar) /2 );
        }
        return heightOffset;
    }

    get localCity(): City {
        if(! globalThis.CITIES ) {
            this.error(`cant get local city if player; CityManager not found`);
            return;
        }
        return globalThis.CITIES.AtPos(this.pos);
    }

    /* PLAYER STORAGE */
    get storage(): TradeNode {
        if (! this.army) { return null; }
        return this.army.storage;
    }

    getStoredRes(resId: string): StoredResource {
        if (! this.storage) { return null; }
        let resInStorage: StoredResource = this.army.getStoredRes(resId);
        return resInStorage;
    }

    amountInStorage(resId: string): number {
        if (! this.storage) { return 0; }
        const fixedStr = this.storage.amountOf(resId).toFixed(1);
        return parseFloat(fixedStr);
    }

    /* EVENT HANDLING */
    onEvent(ev: IBaseEvent) {
        if (ev.id == EV.ID_CITIES_PLACED) {
            // this.createInitialArmy();
        }
    }

    /* PLAYER GETTERS */

    toString(): string {
        return `<Player ${this.pos.x}|${this.pos.z}>`;
    }

    get tile(): WorldTile {
        return globalThis.MAP.tileAtPos(this.pos);
    }

    get scenePos(): BABYLON.Vector3 {
        return WorldTools.worldToScenePos(this.pos);
    }

    static get HEIGHT_OFFSET() { return 0; }
}

/* == CONSTANTS == */
// default properties
export const STARTING_MONEY = 500;
// 3d properties
export const ASSET_WATER = "ship_smallest";
export const ASSET_LAND = "rider";