// common
import { Loggable } from "../../../../common/log";
import { TurnStageNumber } from "../../../../common/turn";
// trade
import { TradeResource } from "./resource";
import { ResourceTypes } from "./resource-types";

/* == INTERFACES == */
export interface IStoredResource {
    resId: string;
    amount: number;
    price: number;
    tradeRes: TradeResource;
}

export interface IStorageEntity {
    storedResources: Array<StoredResource>;
    getStoredRes( resId, cheapest?: boolean ): StoredResource;
}

export interface IResourceStorage {
    maxSpace: number;
    add(resId: string, amount: number): boolean;
    remove(resId: string, amount: number, force?: boolean);
}

/* == STORED RESOURCE == */
export class StoredResource extends Loggable implements IStoredResource {
    resId: string;
    amount: number;
    // optional
    price: number = null;
    profit: number = null;

    constructor(resId: string, amount: number, price?: number, profit?: number) {
        super("StoredRes");
        this.resId = resId;
        this.amount = amount;
        if ( price ) { this.price = price; }
        if ( profit ) { this.profit = profit; }
    }

    get type(): TradeResource {
        if ( ! globalThis.RESOURCE_TYPES ) {
            this.log("cant get resource type; resource types not loaded"); return;
        } else if (! globalThis.TRADE) {
            this.log("cant get resource type; TradeManager not loaded"); return;
        }
        return globalThis.TRADE.getRes(this.resId);
    }

    get tradeRes(): TradeResource { return this.type; }

    toString(): string {
        return `<StoredRes ${this.amount}x${this.resId} price=${this.price}>`
    }

    static fromString(storedResStr): StoredResource {
        const splittedStr = storedResStr.split("x");
        const amount = parseFloat(splittedStr[0]);
        const resId = splittedStr[1];
        return new StoredResource(resId, amount);
    }
}

/* == STORAGE MODEL == */
export class ResourceStorage extends Loggable implements IResourceStorage {
    maxSpace: number;
    spaceTaken: number = 0;
    parent: any;
    resIds: Array<string> = [];
    protected _resAmount = {};
    protected _resPrice = {};

    constructor(maxSpace: number, parent?: any, justFor?: Array<string>) {
        super("ResourceStorage");
        this.maxSpace = maxSpace;
        if ( parent ) { this.parent = parent; }
        if ( justFor && justFor.length > 0 ) {
            this.initializeForTypes(justFor);
        }
        this.initializeEmpty();
    }
    
    /* add "amount" of resources */
    add(resId: string, amount: number): boolean {
        if (this.spaceTaken < this.maxSpace) {
            this._resAmount[resId] += amount;
            this.spaceTaken += amount;
            // this.log(`added ${amount} of ${resId} to storage`);
            return true;
        }
        // this.error(`cant add ${amount} of ${resId} to storage; is already full`);
        return false;
    }

    /* remove "amount" resources without checking whether there are enough left or changing prices */
    remove(resId: string, amount: number, force?: boolean): boolean {
        const storedAmount = this._resAmount[resId];
        if ( storedAmount >= amount) {
            this._resAmount[resId] -= amount;
            // this.log(`removed ${amount} of ${resId} from storage`);
            return true;
        } else {
            // this.error(`cant remove ${amount} of ${resId} from storage; just ${storedAmount} left`);
            return false;
        }
    }

    /* returns how many units of a resource is stored here */
    amountOf(resId: string): number {
        return this._resAmount[resId];
    }

    private initializeEmpty() {
        const allResIds = new ResourceTypes()._resourceIds;
        for (const resId of allResIds) {
            this.resIds.push(resId);
            this._resAmount[resId] = 0;
            const defaultPrice = globalThis.RESOURCE_TYPES.get(resId).defaultPrice;
            this._resPrice[resId] = defaultPrice;
        }
    }

    private initializeForTypes(forIds: Array<string>) {
        for (const resId of forIds) {
            this.resIds.push(resId);
            this._resAmount[resId] = 0;
            const defaultPrice = globalThis.RESOURCE_TYPES.get(resId).defaultPrice;
            this._resPrice[resId] = defaultPrice;
        }
    }

    getStoredResource(resId: string): StoredResource {
        const amount = this.amountOf( resId );
        let stored: StoredResource = new StoredResource( resId, amount);
        return stored;
    }
    
}