// events
import { IBaseEvent, BaseEventListener } from '../../../../common/events';
import * as EV from "../../../game-events";
// world
import { WorldPos } from "../../../world/pos";
// entities
import { Unit } from "../unit";
import { IUniqueInstance } from '../../../../common/entity';
import { ResourceStorage, StoredResource, IStorageEntity } from "../../trade/resource/resource-storage";
import { TradeNode } from "../../trade/trade-node";
import { ICommandableMovableEntity } from '../../../../common/entity';
import { City, CITY_PROP_MIGHT } from "../../city/city";
// helpers
import { RandomElement } from "../../../../helpers/random";
import { IBaseCommand } from '../../../../common/command';
import { OwnedLocation } from '../../../world/location';
import { IUnitType, UnitType } from '../type/unit-type';
import { IFaction } from '../../faction/faction';
import { PlayerInfoDialog } from '../../../gui/dialogs/player-dialog';
import { ArmyCategory } from './army-category';

/* == INTERFACE == */
export interface IUnitArmy extends IStorageEntity, ICommandableMovableEntity {
    name: string;
    armyCat: ArmyCategory;
    home: City;
    // parent
    _ownerKey: string;
    parentCity: City;
    parentFaction: IFaction;
    parentIsPlayer: boolean;
    // army type
    armyTypeIcon: string;
    // flags
    isInField: boolean;
    isFortified: boolean;
    // units
    units: Array<Unit>;
    addUnit(unitObj: Unit);
    addUnitOfType(typeId: string);
    // storage
    storageSpace: number;
    storage: TradeNode;
    storedResources: Array<StoredResource>;
}

/* == ARMY MODEL == */
export class UnitArmy extends OwnedLocation implements IUnitArmy {
    home: City;
    armyCat: ArmyCategory;
    // flags
    isFortified: boolean = false;
    // parent
    _ownerKey: string;
    parentCity: City;
    parentFaction: IFaction;
    parentIsPlayer: boolean = false;
    // commands
    commands: IBaseCommand[];
    // unit list
    units: Array<Unit> = [];
    // storage
    storageSpace: number = 4;
    storage: TradeNode;
    storedResources: Array<StoredResource> = [];

    constructor(armyCatNum: number, pos: WorldPos, overwriteStorageSpace?: number, targetPos?: WorldPos) {
        super(-1, pos, "Unnamed Army", null);
        this.armyCat = new ArmyCategory(armyCatNum);
        // optional parameters
        if ( overwriteStorageSpace ) { this.storageSpace = overwriteStorageSpace; }
        this.storage = new TradeNode( this.storageSpace, true , this);
        // give random name
        this.name = this.getRandomName();
        // fire ArmySpawnedEvent
        const gameStarted: boolean = globalThis.GAME_HAS_STARTED;
        globalThis.EVENTS.fireEvent( new EV.ArmySpawnedEvent(this.num, armyCatNum, pos, -1) );
    }
    
    addUnit(unitObj: Unit): void {
        // set army in Unit model
        unitObj.army = this;
        // add to unit list
        this.units.push(unitObj);
        this.onUnitAdded(unitObj);
    }

    addUnitOfType(typeId: string) {
        const newUnit = globalThis.UNITS.createUnit(typeId, this.pos);
        this.addUnit(newUnit);
    }

    /* ARMY GETTERS */
    get unitsTypeKeys(): Array<string> {
        const unitKeys: Array<string> = [];
        for (const unit of this.units) {
            if (unit) { unitKeys.push(unit.type.key) }
        }
        return unitKeys
    }

    get armyTypeIcon(): string { return this.armyCat.iconName }

    get count(): number { return this.units.length; }

    get length(): number { return this.units.length; }
    get isCaravan(): boolean { return this.armyCat.isCaravan; }

    get isInField(): boolean {
        if (this.armyCat.isMilitia) { return false; } 
        const isInCity = globalThis.CITIES && globalThis.CITIES.getCityAt(this.pos);
        return ! isInCity;

    }

    get freeStorageSpace(): number {
        const totalSpace = this.storageSpace;
        const takenSpace = this.storedResources.length;
        return totalSpace - takenSpace;
    }

    /* return whether this army is of given ARMY_TYPE number */
    isOfType(armyType: number): boolean {
        return this.armyCat.num == armyType;
    }

    get unitTypes(): Array<IUnitType> {
        const unitTypes: Array<IUnitType> = [];
        for (const unit of this.units) {
            if (! unitTypes.includes(unit.type)) {
                unitTypes.push(unit.type);
            }
        }
        return unitTypes;
    }

    get ownerFaction(): IFaction {
        if (! globalThis.FACTIONS) {
            this.warn(`cant get army owner faction: FactionManager not found`); return;
        }
        const factionByKey = globalThis.FACTIONS.getFactionByKey(this._ownerKey);
        return factionByKey;
    }

    private getRandomName(): string {
        // get prefix
        const randomPrefixes = ["Strong", "Zelous", "Loyal", "Free", "Bold", "Quick", "Victorious", "Vigilant", "Blessed"];
        // use city or empire name as prefix for militias, patrols or attack armies
        let prefix: string = RandomElement.of(randomPrefixes);
        // change prefix for player or city owned armies
        if (this.parentIsPlayer) { prefix = "Player"; }
        if (this.parentCity) { prefix = this.parentCity.name; }
        // postfix
        const randomNouns = ["Army", "Force", "Expedition", "Soldiers", "Troops", "Banners", "Spears", "Swords"];
        let postfix: string = RandomElement.of(randomNouns);
        return prefix + " " + postfix;
    }

    /* ARMY STORAGE */
    removeSingleStoredResource(resId: string): void {
        let lastStored: StoredResource = null;
        let lastStoredIndex: number = -1;
        let _index: number = 0;
        for (const storedRes of this.storedResources) {
            if( storedRes.resId == resId ) {
                lastStored = storedRes;
                lastStoredIndex = _index;
            }
            _index += 1;
        }
        if ( lastStored ) {
            this.storedResources.splice(lastStoredIndex, 1);
            this.log(`1 ${resId} removed from storage`);
        }
    }

    /* STORAGE */
    getStoredRes( resId: string, cheapest?: boolean ): StoredResource {
        let ofResType: Array<StoredResource> = [];
        for ( const stored of this.storedResources ) {
            if (stored.resId == resId) { ofResType.push( stored ); }
        }
        if ( ofResType.length < 1 ) {
            this.error(`cant get stored resource ${resId}: none in storage`); return;
        }
        if (! cheapest ) { return ofResType[0]; }
        else {
            let sortedByPrice: Array<StoredResource> = ofResType.sort( function(stored1, stored2) {
                if ( stored1.price > stored2.price ) { return 1; }
                else if ( stored1.price < stored2.price ) { return -1; }
                return 0;
            });
        }
    }

    /* AVERAGE UNIT PROPERTIES */
    get averageDamage(): number {
        let totalDmg: number = 0;
        for (const unit of this.units) {
            const unitBaseDmg = unit.type.gear.weapon.baseAtk;
            totalDmg += unitBaseDmg;
        }
        // average unit damage
        return totalDmg / this.units.length;
    }

    get averageMovement(): number {
        let totalMovement: number = 0;
        for (const unit of this.units) {
            const unitMovement = unit.type.movement;
            totalMovement += unitMovement;
        }
        return totalMovement / this.units.length;
    }

    get averageHealth(): number {
        let totalHp: number = 0;
        for (const unit of this.units) {
            const unitHp = unit.hp;
            totalHp += unitHp;
        }
        return totalHp / this.units.length;
    }

    get armyTypeName(): string { return this.armyCat.name; }

    /* ON EVENTS */
    onUnitAdded(added: Unit) {
        if ( this.parentCity ) {
            // increase city might property
            this.parentCity.might += 1;
        }
    }

    /* EVEN HANDLING */
    onEvent(ev: IBaseEvent) {
        super.onEvent( ev );
    }

    /* COMMANDS */
    onTurn() {
        
    }

    addCommand(cmd: IBaseCommand): void {
        
    }

    onCommand(cmd: IBaseCommand): void {
        
    }

    /* MOVEMENT */
    moveInDir(dirStr: string) {
        
    }

    /* STATIC & TOSTRING */
    toString(): string {
        const unitsStr = this.unitsTypeKeys.join(", ");
        return `<Army "${this.name}" cat=${this.armyCat.name} units=[${unitsStr}]>`;
    }

    static get DEFAULT_MODEL() { return "army_camp"; }
}

/* == ARMY CONSTANTS == */