// common
import { BaseEvent } from "../common/events";
// world
import { WorldPos } from "./world/pos";
// entities
import { City } from "./entities/city/city";
import { ILocation } from "./world/location";

/* == GAME LOADING == */
export class EventManagerStartedEvent extends BaseEvent {
    constructor() {
        super( ID_EVENTS_STARTED );
    }
}

export class AssetsLoadedEvent extends BaseEvent {
    constructor(assetsLoaded: number) {
        super( ID_ASSETS_LOADED );
        this.properties.addInteger("count", assetsLoaded);
    }
}

export class MapFileLoadedEvent extends BaseEvent {
    constructor( mapName: string ) {
        super( ID_MAP_LOADED );
        this.properties.addString( "name", mapName );
    }
}

export class UnitMountsAndGearLoadedEvent extends BaseEvent {
    constructor(totalCount: number) {
        super( ID_MOUNTS_GEAR_LOADED );
        this.properties.addInteger("total", totalCount);
    }
}

export class UnitTypesInitializedEvent extends BaseEvent {
    constructor(unitTypesCount: number) {
        super( ID_UNIT_TYPES_INITIALIZED );
        this.properties.addInteger("count", unitTypesCount);
    }
}

export class CityListLoadedEvent extends BaseEvent {
    constructor( citiesCount: number) {
        super( ID_CITY_LIST_LOADED );
        this.properties.addInteger( "count", citiesCount );
    }
}

export class CitiesPlacedEvent extends BaseEvent {
    constructor( citiesCount: number ) {
        super( ID_CITIES_PLACED );
        this.properties.addInteger( "count", citiesCount );
    }
}

/* == FILES == */
export class OnJsonFileLoadedEvent extends BaseEvent {
    constructor(shortPath: string, fullPath: string, content: {}) {
        super(ID_FILE_JSON_LOADED);
        this.properties.addString("shortPath", shortPath);
        this.properties.addString("fullPath", fullPath);
        // add keys as str array
        const jsonKeys = Object.keys(content);
        this.properties.addStrArray("keys", jsonKeys);
    }
}

export class FileLoadingErrorEvent extends BaseEvent {
    constructor(fullPath: string) {
        super(ID_LOADING_ERROR);
        this.properties.addString("fullPath", fullPath);
    }
}

/* == TURN == */
export class BeforeTurnStartedEvent extends BaseEvent {
    constructor(turnNum: number) {
        super(ID_TURN_BEFORE_STARTED);
        this.properties.addInteger("num", turnNum);
    }
}

export class TurnStartedEvent extends BaseEvent {
    constructor(turnNum: number) {
        super(ID_TURN_STARTED);
        this.properties.addInteger("num", turnNum);
    }

    toString(): string { return `<TurnStarted ${this.properties.getNum("num")}`; }
}

export class PlayerTurnFinishedEvent extends BaseEvent {
    constructor( turnNum: number ) {
        super(ID_TURN_FINISHED);
        this.properties.addInteger("num", turnNum);
    }
}

/* == PLAYER == */
export class PlayerMovedEvent extends BaseEvent {
    constructor(targetPos: WorldPos, directionStr: string, sourcePos?: WorldPos) {
        super(ID_PLAYER_MOVED);
        this.properties.addWorldPos("target", targetPos);
        this.properties.addString("dirStr", directionStr);
        if ( sourcePos ) {
            this.properties.addWorldPos("sourcePos", sourcePos);
        }
    }
}

/* == UNIT & ARMY == */
export class ArmySpawnedEvent extends BaseEvent {
    constructor( armyNum: number, armyType: number, pos: WorldPos, parentNum: number ) {
        super( ID_ARMY_SPAWNED );
        this.properties.addInteger("num", armyNum);
        this.properties.addInteger("armyType", armyType);
        this.properties.addWorldPos("pos", pos);
        this.properties.addInteger("parentNum", parentNum);
        // before game start?
        this.properties.addBoolean("beforeStart", ! globalThis.GAME_HAS_STARTED);
    }
}

/* == CITY == */
export class CityEnteredEvent extends BaseEvent {
    constructor(city: City) {
        super(ID_CITY_ENTERED);
        this.properties.addInteger("num", city.num);
        this.properties.addString("name", city.name);
    }
}

export class CityResourceTradeRequest extends BaseEvent {
    
}

export class ConstructionFinishedEvent extends BaseEvent {
    constructor(location: ILocation, bldId: string, bldLevel: number) {
        super(ID_CONSTR_FINISHED);
        this.properties.addInteger("locationNum", location.num);
        this.properties.addString("bldId", bldId);
        this.properties.addInteger("bldLevel", bldLevel);
    }
}

export class CityPopulationIncreasedEvent extends BaseEvent {
    constructor(city: City, popClassNum: number, newPopulation: number) {
        super(ID_POP_INCREASED);
        this.properties.addInteger("cityNum", city.num);
        this.properties.addInteger("popClass", popClassNum);
        this.properties.addInteger("population", newPopulation);
    }
}

/* == TRADING == */
export class CityDeliveryMadeEvent extends BaseEvent {
    constructor(sourceCity: City, targetCity: City, resId: string, resAmount: number, profit: number) {
        super(ID_TRADE_DELIVERY_MADE);
        this.properties.addInteger("sourceCityNum", sourceCity.num);
        this.properties.addInteger("targetCityNum", targetCity.num);
        this.properties.addString("resource", resId);
        this.properties.addInteger("amount", resAmount);
        this.properties.addInteger("profit", profit);
    }
}

export class PlayerTradedEvent extends BaseEvent {
    constructor(city: City, hasSold: boolean, resId: string, resAmount: number, profit: number, cityPrice: number) {
        super(ID_TRADE_PLAYER);
        this.properties.addInteger("cityNum", city.num);
        this.properties.addBoolean("hasSold", hasSold);
        this.properties.addString("resource", resId);
        this.properties.addInteger("amount", resAmount);
        this.properties.addInteger("profit", profit);
        this.properties.addInteger("price", cityPrice);
    }
}

/* == GUI == */
export class GuiElementClickedEvent extends BaseEvent {
    constructor(elemId: string, isButton?: boolean) {
        super( ID_GUI_ELEM_CLICKED );
        this.properties.addString("elemId", elemId);
        if ( isButton ) {
            this.properties.addBoolean("button", true);
        }
    }
}

export class GuiMenuUpdateRequest extends BaseEvent {
    constructor(elemId: string, partialUpdateData?: string) {
        super(ID_GUI_MENU_UPDATE);
        this.properties.addString("elemId", elemId);
        if ( partialUpdateData ) {
            this.properties.addString("partialUpdateData", partialUpdateData);
        }
    }

}

// please add in correct order
// loading
export const ID_EVENTS_STARTED = "EventManagerStarted";
export const ID_ASSETS_LOADED = "AssetsLoaded";
export const ID_MOUNTS_GEAR_LOADED = "UnitMountAndGearLoaded";
export const ID_UNIT_TYPES_INITIALIZED = "UnitTypesInitialized";
export const ID_MAP_LOADED = "MapFileLoaded";
export const ID_CITY_LIST_LOADED = "CityListLoaded";
export const ID_CITIES_PLACED = "CitiesPlaced"
// files
export const ID_FILE_JSON_LOADED = "JsonFileLoaded";
export const ID_LOADING_ERROR = "FileLoadingError";
// turns
export const ID_TURN_BEFORE_STARTED = "BeforeTurnStarted";
export const ID_TURN_STARTED = "TurnStarted";
export const ID_TURN_FINISHED = "TurnFinished";
// player
export const ID_PLAYER_MOVED = "PlayerMoved";
// units
export const ID_ARMY_SPAWNED = "ArmyCreated";
// city
export const ID_CITY_ENTERED = "CityEntered";
export const ID_CONSTR_FINISHED = "ConstructionFinished";
export const ID_POP_INCREASED = "PopulationIncreased";
// trade
export const ID_TRADE_DELIVERY_MADE = "CityDelivery";
export const ID_TRADE_PLAYER = "PlayerTrade";
// gui
export const ID_GUI_ELEM_CLICKED = "GuiElementClicked";
export const ID_GUI_MENU_UPDATE = "GuiMenuUpdateRequest;"
export const ID_GUI_DIAG_SHOW_REQUEST = "GuiDiagShow";
export const ID_GUI_DIAG_CLOSE_REQUEST = "GuiDiagClose";