// common
import { DawnbringerPalette } from "../../../common/colors";
// events
import { IBaseEvent } from "../../../common/events";
import * as EV from "../../game-events";
// entities
import { City} from "../../entities/city/city";
import { StoredResource } from "../../entities/trade/resource/resource-storage";
// gui basics
import * as GUI_CONST from '../common/constants';
import { EmptyDialog } from "./common/base-dialog";
import { GuiRectangle } from "../common/rectangle";
// gui dialogs
import { ResourceListDialog } from "./reslist-dialog";
import { CityConnectionsDialog } from "./connections-dialog";
import { CityPopulationDialog } from "./city-population-dialog";
import { CityBuildingsDialog } from "./city-buildings-dialog";
import { ArmiesDialog } from "./armies-dialog";
import { RecruitmentDialog } from "./recruitment-dialog";
// babylon
import * as GUI from 'babylonjs-gui';
import { IFactionFlag } from "../../entities/faction/faction-flag";

/* == CITY DIALOG == */
/* Main dialog that is opened on city click (or possibly entering);
   Shows name, owner, money plus clickable datarows for more information to population etc */
export class CityDialog extends EmptyDialog {
    city: City;
    onlyView: boolean;

    constructor(city: City) {
        super(DIAG_CITY, false, city.name, null, false, false, city.ownerEmpire);
        this.city = city;
        this.onlyView = ! this.playerIsInCity;
    }

    get playerIsInCity(): boolean {
        const playerPos = globalThis.PLAYER.pos
        return playerPos && this.city.pos.equals(playerPos);
    }

    get cityIsPlayerOwned(): boolean {
        const playerFamilyKey = globalThis.PLAYER.family.key;
        return this.city.ownerFamily && this.city.ownerFamily.key == playerFamilyKey;
    }

    build(): void {
        super.build();
    }

    addCustomContent() {
        // connections
        const connectionsText = `${this.city.connections.length}`;
        this.addNewDataRow("Connections", connectionsText, "connection", true, DIAG_CITY_LBL_CONNECTIONS);
        // properties
        this.addCityPropertiesBar();
        // owner
        this.addNewDataRow("Owner", this.city.ownerName, "empire", false);
        // leader
        let leaderName: string = "";
        if ( this.city.leader && this.city.leader.fullName ) { leaderName = this.city.leader.fullName; }
        this.addNewDataRow("Leader", leaderName, "trader", false);
        // money
        const moneyText = "" + Math.round(this.city.money);
        this.addNewDataRow("Money", moneyText, "money", false);
        // pops
        const growthIncreaseStr = ` (Grw.: +${this.lastGrowthIncrease.toFixed(1)})`;
        const popsText = Math.round(this.city.population.totalPops) + growthIncreaseStr;
        this.addNewDataRow("Population", popsText, "growth", true, DIAG_CITY_LBL_POPS);
        // buildings
        const buildingsText = "" + this.city.buildings.length;
        this.addNewDataRow("Buildings", buildingsText, "building", true, DIAG_CITY_LBL_BUILDINGS);
        // milita
        const militiaText = this.getMiltiaUnitTypes();
        this.addNewDataRow("Militia", militiaText, "might", true, DIAG_CITY_LBL_MILITIA);
        this.addNewDivider( true );
        // player is currently in this city?
        if (this.playerIsInCity) {
            // res info
            this.addResourceInformation();
            // add trade & recruitment button
            this.addNewButton(DIAG_CITY_BTN_TRADE, "Trade");
            this.addNewButton(DIAG_CITY_BTN_RECRUIT, "Recruit");
        }
        // city is owned by this player's family?
        if ( this.cityIsPlayerOwned ) {
            // plus city command options if city is player owned
            this.addNewButton(DIAG_CITY_BTN_COMMAND, "Command");
        }
    }

    private addCityPropertiesBar() {

    }

    get resHappinessStr(): string {
        const resHappinessPercent = Math.round(this.city.population.lastConsumptionSuccess * 100);
        return `${resHappinessPercent}%`;
    }

    get lastGrowthIncrease(): number { return this.city.population.totalGrowthIncrease; }

    private getMiltiaUnitTypes(): string {
        if (! this.city.militia) {
            return "empty";
        }
        let unitNames: Array<string> = [];
        for (const unit of this.city.militia.units) {
            unitNames.push(unit.type.id);
        }
        return unitNames.join(", ");
    }

    private addResourceInformation() {
        // "Resources:" label
        let resInfoTitle: GUI.TextBlock = this.addNewLabel(`${DIAG_CITY_RESINFO}-title`, "  Resources:", false, true);
        resInfoTitle.textHorizontalAlignment = GUI.TextBlock.HORIZONTAL_ALIGNMENT_LEFT;
        resInfoTitle.fontFamily = GUI_CONST.FONT_HIGHLIGHTED;
        // resource info containter
        const resInfoOutter = new GUI.Rectangle(`${DIAG_CITY_RESINFO}-outter`);
        resInfoOutter.width = GUI_CONST.DEFAULT_INNER_WIDTH;
        resInfoOutter.height = GUI_CONST.DEFAULT_LABEL_HEIGHT;
        resInfoOutter.thickness = 0;
        const resInfoBg = new GUI.Image(`${DIAG_CITY_RESINFO}-bg`, GUI_CONST.BG_PAPER);
        resInfoBg.width = GUI_CONST.DEFAULT_INNER_WIDTH;
        resInfoBg.height = GUI_CONST.DEFAULT_LABEL_HEIGHT;
        resInfoOutter.addControl ( resInfoBg );
        // most resources
        const resInfoRow = new GUI.StackPanel(DIAG_CITY_RESINFO);
        resInfoRow.height = GUI_CONST.DEFAULT_LABEL_HEIGHT;
        resInfoRow.isVertical = false;
        resInfoRow.horizontalAlignment
        resInfoRow.width = GUI_CONST.DEFAULT_INNER_WIDTH;
        // add row to panel
        resInfoOutter.addControl(resInfoRow);
        // well-stocked resources
        for (const storedRes of this.city.tradeNode.getMostStored()) {
            resInfoRow.addControl( this.makeResourceStorageElement(storedRes, true) );
        }
        this.addControl( resInfoOutter, true );
        this.addNewDivider(true);
    }

    private makeResourceStorageElement(storedRes: StoredResource, colorByAmount: boolean): GUI.Rectangle {
        const elemId = `diag-res-max-${storedRes.resId}`;
        // create outter ractangle
        const container = new GUI.Rectangle(elemId);
        container.height = GUI_CONST.DEFAULT_LABEL_HEIGHT;
        container.color = "black";
        container.thickness = 0;
        // container.background = "black";
        const widthInPx = `${Math.floor(400 /3)}px`;
        container.width = widthInPx;
        // get resource label info
        const resObj = globalThis.RESOURCE_TYPES.get(storedRes.resId);
        let colorHexStr: string = BABYLON.Color3.FromArray(resObj.color).toHexString();
        if (colorByAmount) {
            /* when a resource should be colored by its amount in storage,
               paint scarce ones red and well-stocked ones green */
            let amountColorRgb: Array<number> = DawnbringerPalette.DIMGRAY;
            if (storedRes.amount > 15) { amountColorRgb = DawnbringerPalette.DELL; }
            else if (storedRes.amount <= 5) { amountColorRgb = DawnbringerPalette.BROWN; }
            colorHexStr = BABYLON.Color3.FromArray(amountColorRgb).toHexString();
        }
        const amount = Math.round(storedRes.amount);
        const labelText = `${amount} ${resObj.name}`;
        // create resource label
        const label = new GUI.TextBlock(`${elemId}-label`, labelText);
        container.addControl(label);
        label.width = widthInPx;
        label.height = GUI_CONST.DEFAULT_ELEMENT_HEIGHT;
        label.fontSize = GUI_CONST.TEXTSIZE_NORMAL;
        label.fontFamily = GUI_CONST.FONT_HIGHLIGHTED;
        label.color = colorHexStr;
        return container;
    }

    // on child click
    onButtonClick(btnId: string) {
        super.onButtonClick(btnId);
        // resource list btn
        if (btnId == DIAG_CITY_BTN_TRADE) { this.switchToDialog( new ResourceListDialog(this.city.tradeNode) ); }
        // city recruit btn
        else if ( btnId == DIAG_CITY_BTN_RECRUIT ) {
          const playerArmy = globalThis.PLAYER.army;
          const recruitDiag = new RecruitmentDialog(playerArmy);
          globalThis.GUI.showDialog(recruitDiag);
        }
        // city command btn
        else if ( btnId == DIAG_CITY_BTN_COMMAND ) { this.log("COMMAND"); }
    }

    onElementClicked(elemId: string): boolean {
        const success = super.onElementClicked(elemId);
        const city = this.city;
        if (! success ) { return false; }
        // buildings
        if (elemId == 'datarow-Buildings') { this.switchToDialog( new CityBuildingsDialog(city) ); }
        // connections
        else if (elemId == 'datarow-Connections') { this.switchToDialog( new CityConnectionsDialog(city) ); }
        // militia
        else if (elemId == 'datarow-Militia') { this.switchToDialog( new ArmiesDialog(this.city.pos) ); }
        // pops
        else if (elemId == 'datarow-Population') { this.switchToDialog( new CityPopulationDialog(city) ); }
        // else
        else { this.log(`element ${elemId} clicked but no action is connected`); }
        return true;
    }

    // EVENT LISTENING
    onEvent(ev: IBaseEvent) {
        super.onEvent(ev);
        // ELEMENT CLICKED
        if ( ev.id = EV.ID_GUI_ELEM_CLICKED) {
            let clickedElemId: string = ev.properties.getStr("elemId");
            if (clickedElemId) { this.onElementClicked(clickedElemId); }
        }
    }
}


/* == CONSTANTS == */
// city dialog
export const DIAG_CITY = "diag-city";
export const DIAG_CITY_RESINFO = "diag-city-resinfo";
export const DIAG_CITY_BTN_TRADE = "diag-city-btn-trade";
export const DIAG_CITY_BTN_RECRUIT = "diag-city-btn-recruit";
export const DIAG_CITY_BTN_COMMAND = "diag-city-btn-command";
export const DIAG_CITY_LBL_LEADER = "diag-city-leader";
export const DIAG_CITY_LBL_CONNECTIONS = "diag-city-connections";
export const DIAG_CITY_LBL_MILITIA = "diag-city-militia";
export const DIAG_CITY_LBL_BUILDINGS = "diag-city-buildings";
export const DIAG_CITY_LBL_POPS = "diag-city-growth";
