// events
import { IBaseEvent, BaseEvent, BaseEventListener } from "../../../../common/events";
import * as EV from "../../../game-events";
// entities
import { IFaction } from '../../../entities/faction/faction';
import { FactionFlag } from '../../../entities/faction/faction-flag';
// gui
import * as GUI_CONST from '../../common/constants';
import { IGuiContainer, FlexGuiLayout } from '../../common/layouts';
import { DialogButton } from './dialog-button';
// gui
import * as GUI_EV from "../../common/gui-events";
import { GuiRectangle } from '../../common/rectangle';
import { FactionFlagImage } from "../../elements/babylon/flags";
// babylon
import * as GUI from 'babylonjs-gui';
import { GuiScrollView } from "../../elements/scroller";

/* == DIALOG INTERFACE == */
export interface IDialog extends IGuiContainer {
    // basic properties
    isBlocking: boolean;
    // dialog flags
    hasCustomBackground: boolean;
    buttonIds: Array<string>;
    showFactionFlag: IFaction;
    // dialog methods
    close();
    switchToDialog(dialog: IDialog);
    // dialog event listeners
    onButtonClick(btnId: string);
    onElementClicked(elemId: string): boolean;
    onSwitcherChange(elemId: string, selected: string);
    
}

/* == EMPTY DIALOG == */
export class EmptyDialog extends FlexGuiLayout implements IDialog {
    titleText: string = "Unnamed Dialog";
    // flags
    hasCustomBackground: boolean = false;
    isBlocking: boolean;
    showFactionFlag: IFaction = null;
    // child elements
    inner: GuiRectangle;
    buttonIds: Array<string> = ["cancel", "ok"];
    noButtonDivider: boolean = false;

    constructor(id: string, isBlocking: boolean, headerText?: string, extraButtons?: Array<string>, 
        customBackground?: boolean, noButtonDivider?: boolean, showFactionFlag?: IFaction) {
        super(id, GUI_CONST.DIR_VERT, headerText);
        this.isBlocking = isBlocking;
        // optionals
        if (customBackground) { this.hasCustomBackground = true; }
        if (noButtonDivider) { this.noButtonDivider = true; }
        if (showFactionFlag) { this.showFactionFlag = showFactionFlag; }
        // add buttons
        this.addButtons(extraButtons);
    }

    switchToDialog(dialog: IDialog) {
        this.close();
        globalThis.GUI.showDialog(dialog, true);
    }

    build(): void {
        // build FlexGuiLayout & set style & sizing
        super.build( false );
        this.outter.height = "70%";
        GuiScrollView.setDefaultStyle(this.outter);
        // build optional children: background image & faction flag
        if (this.hasCustomBackground) { this.buildBackground(); }
        if (this.showFactionFlag) { this.addFactionFlag(); }
        // build button bar
        this.buildButtonBar();
    }

    private addFactionFlag() {
        const factionFlag: FactionFlag = this.showFactionFlag.flag;
        if (factionFlag) {
            const flag = new FactionFlagImage(factionFlag);
            flag.left = "-20px";
            this.outter.addControl(flag);
        }
    }

    private addButtons(newButtonIds: Array<string>): void {
        this.buttonIds = this.buttonIds.concat(newButtonIds);
    }

    addNewDivider(empty?: boolean, setParent?: GUI.Container): GUI.TextBlock {
        const divider = super.addNewDivider(empty, setParent);
        divider.widthInPixels = 400;
        return divider;
    }

    private buildBackground() {
        const outterElemId = `${this.id}-inner`;
        this.inner = new GuiRectangle(outterElemId);
        this.inner.width = GUI_CONST.DIALOG_OUTTER_WIDTH;
        this.inner.height = GUI_CONST.DIALOG_HEIGHT;
        this.addControl(this.inner);
    }

    buildButtonBar() {
        if (! this.noButtonDivider) { this.addNewDivider(true); }
        // add close button
        this.addNewButton(ID_BUTTON_CLOSE, "Close");
    }

    close() {
        this.log(`closing ..`);
        this.outter.dispose();
        this.babylon.dispose();
        globalThis.DIALOG = null;
    }

    onButtonClick(btnId: string) {
        this.log(`button ${btnId} clicked`);
        globalThis.EVENTS.fireEvent ( new EV.GuiElementClickedEvent( btnId, true ));
        if ( btnId == ID_BUTTON_CLOSE) {
            this.close();
        }
    }

    onSwitcherChange(elemId: string, selected: string) {
        this.log(`switcher "${elemId}" changed to ${selected}`);
    }

    addNewButton(btnId: string, text: string): GUI.Button { 
        return DialogButton.CreateNew(btnId, text, this); 
    }

    public onEvent(ev: IBaseEvent): void {
        super.onEvent(ev);
        if (ev.id == GUI_EV.ID_SWITCHER_CHANGE) {
            const elemId = ev.properties.getStr("elemId");
            const selected = ev.properties.getStr("selected");
            this.onSwitcherChange(elemId, selected);
        }
    }

    /* TO STRING */
    toString(): string { 
        const directChildCount = this.children.length;
        return `<Dialog id="${this.id}" title="${this.title}" children=${directChildCount}>`;
    }
}

/* == SPECIAL DIALOG: INFO == */
/* == DIALOG DEFAULT CONSTANTS == */
export const ID_BUTTON_CLOSE = "diag-close";
export const BUTTON_TEXT_COLOR = "ATLANTIS";

/* == SPECIAL DIALOGS CONSTANTS == */