// common
import { RandomDice, RandomElement, RandomInt, RandomWorldPos } from "../../helpers/random";
// events
import { IBaseEvent, BaseEventListener } from '../../common/events';
import * as EV from "../game-events";
// entities
import { UnitArmy } from "../entities/unit/army/army";
import { ArmyCategory } from "../entities/unit/army/army-category";
import { City } from "../entities/city/city";
// manager
import { AbstractManager } from "./abstract-manager";
import { WorldPos } from "../world/pos";
import { IFaction } from "../entities/faction/faction";
import { IEmpire } from "../entities/faction/empire";

export class ArmyManager extends AbstractManager {
    armies: Array<UnitArmy> = [];
    _armyTypes: Array<number> = [];
    _playerArmy: UnitArmy;
    // flags
    hasPlayerArmyCreated: boolean = false;
    hasMilitiasCreated: boolean = false;
    hasEmpireArmiesCreated: boolean = false;
    
    constructor() {
        super("Army");
    }

    /* ARMY CREATION */
    createNewArmy(armyType: number, pos: WorldPos, armyOwnerKey?: string): UnitArmy {
        const army: UnitArmy = new UnitArmy(armyType, pos);
        army.num = this.armies.length;
        if (armyOwnerKey) { army._ownerKey = armyOwnerKey; }
        return this._addArmy(army);
    }

    _addArmy(armyObj: UnitArmy): UnitArmy {
        this.armies.push(armyObj);
        this._armyTypes.push(armyObj.armyCat.num);
        return armyObj;
    }

    /* ARMY GETTERS */
    getArmyByNum(armyNum: number): UnitArmy {
        const armyWithNum = this.armies[armyNum];
        if (! armyWithNum) { this.warn(`cant find army with num ${armyNum}`); }
        return armyWithNum;
    }

    getArmiesAtPos(worldPos: WorldPos): Array<UnitArmy> {
        let armiesAtPos: Array<UnitArmy> = [];
        for (const army of this.armies) {
            if (army.pos.equals(worldPos)) { armiesAtPos.push(army); }
        }
        return armiesAtPos;
    }

    getArmiesByFactionKey(factionKey: string): Array<UnitArmy> {
        const factionArmies: Array<UnitArmy> = [];
        for (const army of this.armies) {
            if (army._ownerKey == factionKey) {
                factionArmies.push(army);
            }
        }
        return factionArmies;
    }

    getArmiesByFaction(faction: IFaction): Array<UnitArmy> {
        return this.getArmiesByFactionKey(faction.key);
    }

    getOutsideArmies(): Array<UnitArmy> {
        const outsideArmies: Array<UnitArmy> = [];
        for (const army of this.armies) {
            if (army.isInField) { outsideArmies.push(army); }
        }
        return outsideArmies;
    }

    static FilterByArmyCatNum(armies: Array<UnitArmy>, armyCatNum: number): Array<UnitArmy> {
        const armiesOfCat: Array<UnitArmy> = [];
        for (const army of armies) {
            if (army.armyCat.num == armyCatNum) { armiesOfCat.push(army); }
        }
        return armiesOfCat;
    }

    filterByArmyCatNum(armies: Array<UnitArmy>, armyCatNum: number): Array<UnitArmy> {
        return ArmyManager.FilterByArmyCatNum(armies, armyCatNum);
    }

    /* shortforms */
    ByNum(num: number): UnitArmy { return this.getArmyByNum(num); }
    AtPos(pos: WorldPos): Array<UnitArmy> { return this.getArmiesAtPos(pos); }

    /* CREATE NEW ARMIES */

    /* create a new army in a city (can be militia or patrol) */
    createNewCityArmy(city: City, isPatrol?: boolean): UnitArmy {
        const ownerFactionKey = city._factionKey;
        let armyCatNum: number = ArmyCategory.CITY_MILITIA;
        if (isPatrol && isPatrol == true) { armyCatNum = ArmyCategory.CITY_PATROL; }
        const army = this.createNewArmy(armyCatNum, city.pos, ownerFactionKey);
        army.parentCity = city;
        army.home = city;
        if (! isPatrol) { city.militia = army;}
        return army;
    }

    createPlayerArmy(): UnitArmy {
        this.log(`creating player starting army`);
        // abort when player has already an army created
        if (this.hasPlayerArmyCreated) { return; }
        // create new UnitArmy object
        const playerFamilyKey = globalThis.PLAYER.family.key;
        this._playerArmy = this.createNewArmy(ArmyCategory.CARAVAN_PLAYER, globalThis.PLAYER.pos, playerFamilyKey);
        // add to player
        globalThis.PLAYER.setArmy(this._playerArmy);
        this._playerArmy.parentIsPlayer = true;
        // add donkey starting unit
        this._playerArmy.addUnitOfType("DN");
        // flag player army creation as done
        this.hasPlayerArmyCreated = true;
        return this._playerArmy;
    }

    createNewFactionArmy(faction: IFaction, startPos?: WorldPos, isPatrol?: boolean): UnitArmy {
        let armyCatNum: number = ArmyCategory.FACTION_ARMY;
        if (isPatrol) { armyCatNum = ArmyCategory.FACTION_PATROL; }
        const factionKey = faction.key;
        const factionArmy = this.createNewArmy(armyCatNum, startPos, factionKey);
        // add to faction
        factionArmy.parentFaction = faction;
        faction.factionArmies.push(factionArmy);
        return factionArmy;
    }

    /* STARTING ARMY CREATION */
    createStartingArmies() {
        // already created all starting armies? abort
        if ( this.checkAllArmiesCreated ) { return; }
        this.log(`creating starting armies..`);
        // create city miltia armies
        if(! this.hasMilitiasCreated) {
            this._createCityMilitias();
        }
        // create empire armies
        if (! this.hasEmpireArmiesCreated) {
            this._createEmpireArmies();
        }
    }

    addRndMilitaryUnitTo(army: UnitArmy) {
        const unitTypeId = this.getRandomMilityUnitTypeId();
        army.addUnitOfType(unitTypeId);
    }

    private getRandomMilityUnitTypeId(): string {
        const militaryTypes = globalThis.UNIT_TYPES.getMilitaryTypes();
        return RandomElement.of(militaryTypes).id;
    }

    private createRndCityMilitia(city: City): UnitArmy {
        if (this.hasMilitiasCreated) { return; }
        const army = this.createNewCityArmy(city, false);
        const startingUnitCount = RandomElement.of([1, 2, 3]);
        for (let unitNum = 0; unitNum < startingUnitCount; unitNum++) {
            this.addRndMilitaryUnitTo(army);
        }
        return army;
    }

    private _createCityMilitias(): Array<UnitArmy> {
        const citiesLoaded = globalThis.CITIES && globalThis.CITIES.getAll().length > 0;
        const militiaArmies: Array<UnitArmy> = [];
        if (! citiesLoaded) {
            this.warn(`cant create city militias; CityManager not initialized`); return;
        }
        const citiesCount = globalThis.CITIES.getAll().length;
        const firstMilitiaArmyNum = this.armies.length;
        for (const city of globalThis.CITIES.getAll()) {
            const militiaArmy = this.createRndCityMilitia(city);
            militiaArmies.push(militiaArmy);
        }
        const militiaArmiesCount = this.armies.length - firstMilitiaArmyNum;
        this.log(`${militiaArmiesCount} for ${citiesCount} cities created`);
        this.hasMilitiasCreated = true;
        return militiaArmies;
    }

    private _createEmpireArmies(): boolean {
        // check for FactionManager
        if (! globalThis.FACTIONS) {
            this.warn(`cant create empire armies: FactionManager not found`); return false;
        }
        const firstEmpireArmyNum = this.armies.length;
        for (const empire of globalThis.FACTIONS.empires) {
            this._createEmpireStartingArmies(empire);
        }
        const empireArmiesCount = this.armies.length - firstEmpireArmyNum;
        this.log(`created ${empireArmiesCount} empire armies`);
        this.hasEmpireArmiesCreated = true;
        return true;
    }

    private _createEmpireStartingArmies(empire: IEmpire): Array<UnitArmy> {
        let startingArmiesCount: number = RandomInt.below(3);
        const startingArmies: Array<UnitArmy> = [];
        for (let empireArmyNum = 0; empireArmyNum <= startingArmiesCount; empireArmyNum++) {
            let isPatrol: boolean = RandomDice.roll(2);
            const empireArmy = this._createSingleEmpireArmy(empire, isPatrol);
            startingArmies.push(empireArmy);
        }
        return startingArmies;
    }

    private _createSingleEmpireArmy(empire: IEmpire, isPatrol: boolean): UnitArmy {
        const army = this.createNewFactionArmy(empire, null, isPatrol);
        // empire has capital? set as home
        const empireCapital = empire.capitalCity;
        if (! empireCapital) {
            this.warn(`creating empire "${empire.key}" army: no home assignable`); }
        else { 
            army.home = empireCapital;
            // set rnd start pos
            const rndStartPos = this.findStartingPosAround(empireCapital.pos, 4);
            army.pos = rndStartPos;
        }
        // add 1 or 2 random military units
        let startingUnitsCount: number = RandomInt.below(3);
        for (let empireUnitNum = 0; empireUnitNum < startingUnitsCount; empireUnitNum++) {
            this.addRndMilitaryUnitTo(army);
        }
        return army;
    }

    public findStartingPosAround(centerPos: WorldPos, radius): WorldPos {
        let isValid: boolean = false;
        while (! isValid) {
            let pos: WorldPos = RandomWorldPos.around(centerPos, radius);
            const tile = globalThis.MAP.tileAtPos(pos);
            if (tile.isOnLand) {
                isValid = true;
                return pos;
            }
        }
        return
    }

    /* ARMY MANAGER GETTERS */
    get checkAllArmiesCreated(): boolean {
        const allArmiesCreated = this.hasPlayerArmyCreated && this.hasMilitiasCreated && this.hasEmpireArmiesCreated;
        if (allArmiesCreated) {
            this.onAllArmiesCreated();
            return true;
        }
        return false;
    }

    /* INTERNAL EVENTS */
    onCityListCreated() {
        this.createPlayerArmy();
    }

    onEmpireCapitalsAssigned() {
        this.createStartingArmies();
    }

    onAllArmiesCreated() {
        this.log(`finished creation of ${this.armies.length} armies for player, cities & faction`);
    }

    /* EVENT HANDLING */
    onEvent(ev: IBaseEvent) {
        super.onEvent(ev);
        if (ev.id == EV.ID_UNIT_TYPES_INITIALIZED) {
            this.createStartingArmies();
        } else if (ev.id == EV.ID_PLAYER_MOVED) {
            this.createStartingArmies();
        }
    }

    get hasLoaded(): boolean { return this.armies.length > 1; }
}