// common
import { RandomElement } from "../../helpers/random";
// events
import { IBaseEvent, BaseEventListener } from '../../common/events';
import * as EV from "../game-events";
// world
import { WorldPos } from "../world/pos";
import { WorldTile } from "../world/tile/tile";
// models
import { City } from "../entities/city/city";
import { AbstractBuildingType } from "../entities/building/type/building-type";
import { IBuildingInstance } from "../entities/building/building";
// babylon
import * as BABYLON from 'babylonjs';
// manager
import { AbstractManager } from "./abstract-manager";
import {IConstructable} from "../entities/building/type/constructable";

export class CityManager extends AbstractManager {
    _names: Array<string> = [];
    cities: Array<City> = [];
    _placedCities: Array<City> = [];
    bldTypes: Array<IBuildingInstance> = [];

    constructor(generate?: boolean) { 
        super("City");
    }

    /* GET & SET CITIES */
    addCity(city: City): void {
        if (! city.num) {
            city.num = this.cities.length;
        }
        this._names.push(city.name);
        this.cities.push(city);
    }

    createCity(name: string, pos: WorldPos, culture: string, primaryResource: string) {
        let newCity: City = new City(null, name, pos, culture, primaryResource);
        this.addCity(newCity);
    }

    getAll(): Array<City> {
        return this.cities;
    }

    get citiesCount(): number {
        if (this.cities && this.cities.length > 0) { return this.cities.length; }
        return 0;
    }

    /* GET BY PROPERTY */
    getByNum(num: number): City {
        const city = this.cities[num];
        return city;
    }

    getByKey(key: string): City {
        for (const city of this.cities) {
            if (city.key == key) { return city; }
        }
        return null;
    }

    getByName(name: string): City {
        let cityIndex: number = this._names.indexOf(name);
        return this.cities[cityIndex];
    }

    getCityAt(worldPos: WorldPos): City {
        for (const city of this.cities) {
            if (city.pos.equals(worldPos)) { return city; }
        }
        return null;
    }

    getCitiesByFactionKey(factionKey: string): Array<City> {
        const factionCities: Array<City> = [];
        for (const city of this.getAll()) {
            if (city._factionKey && city._factionKey == factionKey) {
                factionCities.push(city);
            }
        }
        return factionCities;
    }

    // shortforms
    ByNum(num: number): City { return this.getByNum(num); }
    ByKey(key: string): City { return this.getByKey(key); }
    ByName(name: string): City { return this.getByName(name); }
    AtPos(pos: WorldPos): City { return this.getCityAt(pos); }
    OfOwner(factionKey: string): Array<City> { return this.getCitiesByFactionKey(factionKey); }

    /* CITY MESH */
    getMeshes(): Array<BABYLON.AbstractMesh> {
        let cityMeshes: Array<BABYLON.AbstractMesh> = [];
        for (const city of this.cities) {
            if ( city.mesh ) {
                cityMeshes.push( city.mesh );
            }
        }
        return cityMeshes;
    }

    getRandom(): City {
        return RandomElement.of(this.cities) as City;
    }

    /* public load(): number {
        let cityAmount: number = 0;
        return cityAmount;
    } */

    _onTurnFinished() {
        this.log(`tick`);
        for (const city of this.cities) {
            city.onTurn();
        }
    }

    static get RANDOM_AMOUNT(): number {
        return 20;
    }

    /* notifies a tile that a city is standing on it */
    linkCityTiles() {
        for (const city of this.getAll()) {
            let tileAt: WorldTile = globalThis.MAP.tileAtPos(city.pos);
            tileAt.city = city;
        }
        this.log("flagged city tiles");
    }

    makeCityRegions() {
        // link neighboring tiles
        this.log('TODO generate city regions');
    }

    initializeCityEmpire() {
        let ownedCitiesAmount: number = 0;
        for (const city of this.getAll()) {
            if ( city._factionKey ) { // when city has an owner..
                const empireObj = globalThis.FACTIONS.getEmpireByKey( city._factionKey );
                if (empireObj) {
                    // .. that is an empire
                    empireObj.giveCityOwnership(city, true);
                    ownedCitiesAmount += 1;
                }
            }
        }
        this.log(`given empire ownership to ${ownedCitiesAmount} cities`);
    }

    /* CITY FEATURES */
    initializeCityFeatures() {
        for (const city of this.getAll()) {
            city.loadFeatures();
        }
    }

    /* BUILDINGS */
    getAllBuildingTypes(): IConstructable[] {
        return globalThis.CONSTRUCTABLE.getAll();
    }

    getBuildingTypesByCat(bldCatNum: number): AbstractBuildingType[] {
        let catBuildings = [];
        for (const bldType of this.getAllBuildingTypes()) {
            if (bldCatNum == bldType.category.num) {
                catBuildings.push(bldType);
            }
        }
        return catBuildings;
    }

    /* ON EVENTS */
    onEvent(ev: IBaseEvent): void {
        super.onEvent( ev );
        // when cities are placed -> link to tile
        if ( ev.id == EV.ID_CITIES_PLACED ) {
            this.linkCityTiles();
        }
    }

    /* fired by world scene after all cities have been added */
    onCitiesLoaded() {
        // load city features
        this.initializeCityFeatures();
        // initialize cities ownership
        this.initializeCityEmpire();
        globalThis.ARMIES.onCityListCreated();
        globalThis.FACTIONS.onCitiesLoaded();
        // initialize city regions
        this.makeCityRegions();
        // initialize city connections
        globalThis.CONN.makeCityConnections();
        // fire event after loading features and owner faction
        // globalThis.EVENTS.fireEvent( new EV.CityListLoadedEvent( citiesCount ), true);
        // inform trafe & unit manager
        globalThis.TRADE.onCitiesLoaded();
        globalThis.UNITS.onCitiesLoaded();
        // globalThis.CHARS.createCityLeaders();
    }

}