// common
import { GameOptions } from "../game-options";
import { RandomElement, RandomInt, RandomDice } from "../../helpers/random";
// events
import { IBaseEvent, BaseEventListener } from '../../common/events';
import * as EV from "../game-events";
// entities
import { IFaction } from "../entities/faction/faction";
import { Empire, IEmpire } from "../entities/faction/empire";
import { Family, IFamily } from "../entities/faction/family";
import { Cultures } from "../entities/faction/culture";
import { City } from "../entities/city/city";
// data
import { CustomFamiliesLoader, WorldEmpireLoader } from "./world-loaders";
// manager
import { AbstractManager } from "./abstract-manager";
import { UnitArmy } from "../entities/unit/army/army";
import { ArmyManager } from "./army-manager";

/* == FACTION MANAGER == */
export class FactionManager extends AbstractManager {
    // families
    families: Array<Family> = [];
    playerFamily: Family = null;
    // empires
    empires: Array<Empire> = [];
    _empireKeys: Array<string> = [];
    cultures: Cultures = null;
    // flags
    hasFamilyHomesLoaded: boolean = false;
    hasStartingArmiesCreated: boolean = false;
    hasEmpireCapitalsAssigned: boolean = false;
    // private
    private lastFactionNum: number = -1;
    private _empireCapitalKeys: {} = {};
    private _empireCapitalsAssignedCount: number = 0;

    constructor() {
        super("Faction");
        this.loadCultures();
        this.createDefaultEmpires();
        this.loadEmpiresFromTsFile();
        this.loadOrGenerateFamilies();
    }

    loadOrGenerateFamilies(): Array<IFamily> {
        const customFamilies = new CustomFamiliesLoader().load();
        if (customFamilies && customFamilies.length > 0) {
            let isFirstFamily: boolean = true;
            for (const family of customFamilies) {
                if (isFirstFamily) {
                    family.isAi = false;
                }
                this.addFamily(family as Family);
                isFirstFamily = false;
            }
        }
        else { this.createRandomFamilies(); }
        return null;
    }

    /* FACTION GETTERS */
    getFactionByNum(factionNum: number): IFaction {
        if ( factionNum > this.lastFactionNum ) {
            this.error(`cant get faction with num ${factionNum}; only ${this.lastFactionNum} factions loaded`); return;
        }
        for (const faction of this.allFactions) {
            if (faction.num == factionNum) { return faction; }
        }
        return null;
    }

    getEmpireByKey(empireKey: string): Empire {
        for (const empire of this.empires) {
            if ( empire.key == empireKey ) { return empire; }
        }
        return null;
    }

    get allFactions(): Array<IFaction> {
        let allFactions: Array<IFaction> = [];
        // collect families
        for (const family of this.families) { allFactions.push(family); }
        // .. & empires
        for (const empire of this.empires) { allFactions.push( empire ); }
        // and return all as IFaction array
        return allFactions;
    }

    getFactionByKey(factionKey: string): IFaction {
        // family?
        if (factionKey.startsWith("F") && factionKey.length == 2) {
            const familyNum = parseInt(factionKey.replace("F", ""));
            return this.getFamilyByNum(familyNum);
        }
        // empire?
        return this.getEmpireByKey(factionKey);
    }

    /* shortforms */
    ByKey(key: string) { return this.getFactionByKey(key); }

    /* FAMILY MANAGEMENT */
    addFamily(family: Family): void {
        const familyNum = this.families.length;
        family.setNum(familyNum);
        this.families.push(family);
        if (! family.isAi) { this.playerFamily = family; }
    }

    addRandomFamily(name?: string, randomHome?: boolean): Family {
        const playersAmount = this.families.length
        const isPlayer = playersAmount == 0;
        let rndFamily: Family = Family.random();
        if ( randomHome ) { rndFamily.hasRandomHome = true; }
        if ( isPlayer ) {
            rndFamily.isAi = false;
            this.playerFamily = rndFamily;
        }
        this.addFamily(rndFamily);
        return rndFamily;
    }

    createRandomFamilies(): Array<Family> {
        let families: Array<Family> = [];
        for (let familyNum = 0; familyNum < RND_FAMILY_COUNT; familyNum++) {
            this.addRandomFamily(null, true);
        }
        return families;
    }

    getPlayerFamily(): Family {
        if ( this.families.length == 0 ) {
            this.error(`cant get player faction; families not generated`); return;
        }
        return this.families[0];
    }

    getFamilyByNum(num: number): Family {
        const foundFamily = this.families[num];
        if (! foundFamily) {
            this.error( `no family ${num} found in ${ this.getFamilyNums() }`); return;
        }
        return foundFamily;
    }

    getFamilyNums(): Array<number> {
        let nums: Array<number> = [];
        for (const family of this.families) { nums.push(family.num); }
        return nums;
    }

    getFactionArmies(faction: IFaction): Array<UnitArmy> {
        if (! globalThis.ARMIES) {
            this.warn(`cant get armies of ${faction.name}: ArmyManager not found`); return;
        }
        return globalThis.ARMIES.getArmiesByFaction(faction);
    }

    getFactionArmiesByCat(faction: IFaction, armyCatNum: number) {
        const factionArmies = this.getFactionArmies(faction);
        const factionArmiesOfCat = ArmyManager.FilterByArmyCatNum(factionArmies, armyCatNum);
        return factionArmiesOfCat;
    }

    /* EMPIRE MANAGEMENT */
    addEmpire(empire: Empire): void {
        this._empireKeys.push(empire.key);
        this.empires.push(empire);
    }

    getEmpireCities( empireKey: string ): Array<City> {
        let empireCities: Array<City> = [];
        for (const city of globalThis.CITIES.getAll()) {
            if ( city._factionKey && city._factionKey == empireKey ) { empireCities.push( city ); }
        }
        return empireCities;
    }

    getEmpires(hideDefault?: boolean): Array<IEmpire> {
        const empires: Array<IEmpire> = [];
        for (const empire of this.empires) {
            let hideEmpire: boolean = false;
            if (hideDefault == true && empire.isDefault) { hideEmpire = true; }
            if (! hideEmpire) { empires.push(empire); }
        }
        return empires;
    }

    getEmpireKeys(hideDefault?: boolean): Array<string> {
        const empireKeys: Array<string> = [];
        for ( const nonHiddenEmpire of this.getEmpires(hideDefault) ) {
            empireKeys.push(nonHiddenEmpire.key)
        }
        return empireKeys;
    }

    /* LOADING */
    private loadCultures() {
        this.cultures = new Cultures();
        this.log(`loaded ${this.cultures.length} cultures`);
    }

    private createDefaultEmpires(): void {
        // indie empire
        const independent = new Empire("I", "Independent", null, null);
        this.addEmpire(independent);
    }

    private assignFamilyHomes() {
        if (this.hasFamilyHomesLoaded) { return; }
        for (const family of this.families) {
            const familyHasNoHome: boolean = ! family.home;
            if (familyHasNoHome && family.hasRandomHome) {
                let rndCity: City = globalThis.CITIES.getRandom();
                family.setHome(rndCity);
            }
        }
        this.hasFamilyHomesLoaded = true;
    }

    private loadEmpiresFromTsFile() {
        const mapEmpires = new WorldEmpireLoader().load();
        const empiresCount = mapEmpires.length;
        let empireNames: Array<string> = [];
        for (const empire of mapEmpires) {
            this.addEmpire(empire as Empire);
            empireNames.push(empire.name);
        }
        const empiresStr = empireNames.join(", ");
        this.log(`${empiresCount} empires were loaded: ${empiresStr}`);
    }

    assignEmpireCapitals(loadOnly?: boolean) {
        if (this.hasEmpireCapitalsAssigned) { return; }
        this.log(`start empire capital assignment`);
        // check for FactionManager
        if (! ( globalThis.FACTIONS && globalThis.FACTIONS.empires && globalThis.FACTIONS.empires.length > 1) ) {
            this.warn(`cant assign empire capitals: FactionManager not initialized`); return; }
        let empireCapitalsCount: number = 0;
        for ( const city of globalThis.CITIES.getAll() ) {
            const ownerKey = city._factionKey;
            const empiresWithCapital = Object.keys(this._empireCapitalKeys);
            const ownerHasCapital = empiresWithCapital.includes(ownerKey);
            if (! ownerHasCapital) {
                this.assignSingleEmpireCapital(ownerKey, city.key, true);
            }
        }
        // when all capitals are assigned
        const empireListStr = Object.keys(this._empireCapitalKeys).join(", ");
        this.log(`capitals for ${empireCapitalsCount} empires assigned: ${empireListStr}`);
        this.hasEmpireCapitalsAssigned = true;
        // inform other managers
        globalThis.ARMIES.onEmpireCapitalsAssigned();
    }

    assignSingleEmpireCapital(empireKey: string, capitalCityKey: string, loadOnly?: boolean): boolean {
        this._empireCapitalKeys[empireKey] = capitalCityKey;
        const empireObj: IEmpire = this.getEmpireByKey(empireKey);
        const setToFaction = empireObj && ! loadOnly;
        if (setToFaction) {
            const cityObj = globalThis.CITIES.getByKey(capitalCityKey);
            if (cityObj) { 
                empireObj.home = cityObj;
                this._empireCapitalsAssignedCount += 1;
                return true;
            }
        }
        return false;
    }

    getEmpireCapitalKey(empireKey): string {
        const empiresWithCapitalKeys = Object.keys(this._empireCapitalKeys);
        if (! empiresWithCapitalKeys.includes(empireKey) ) { return; }
        return this._empireCapitalKeys[empireKey];
    }

    /* STARTING ARMIES */
    private createStartingArmies() {
        if (this.hasStartingArmiesCreated) { return; }
        this.log(`placing starting armies for ${this.allFactions.length} factions..`);
        let armiesCreated: number = 0;
        for (const faction of this.allFactions) {
            let amount: number = RandomInt.below( MAX_STARTING_ARMIES );
            for (let index = 0; index < amount; index++) {
                let rndIsPatrol: boolean = RandomDice.roll(2);
                faction.createNewArmy( rndIsPatrol )
                armiesCreated += 1;
            }
        }
        this.hasStartingArmiesCreated = true;
        this.log(`created ${armiesCreated} starting armies`);
    }

    nextFactionNum(): number {
        this.lastFactionNum += 1;
        return this.lastFactionNum;
    }

    /* INTERNAL ON EVENT METHODS */
    onCitiesLoaded() {
        // assign family home & empire capitals
        this.assignFamilyHomes();
        this.assignEmpireCapitals();
        // place starting armies
        this.createStartingArmies();
    }

    onNameListLoaded() {
        this.assignFamilyHomes();
    }

    onUnitTypesInitialized() {
        // this.assignEmpireCapitals();
        // this.createStartingArmies();
    }

    /* GAME EVENT LISTENING */
    onEvent(ev: IBaseEvent): void {
        super.onEvent( ev );
        // gear initialized?
    }

}

/* == CONSTANTS == */
const MAX_STARTING_ARMIES = 4;
const RND_FAMILY_COUNT = 5;