// events
import { BaseEventListener, IBaseEvent } from "../../common/events";
import * as EV from "../game-events";
// world
import { WorldSelection } from "../world/selection";
import { WorldTile } from "../world/tile/tile";
import { WorldPos } from "../world/pos";
import { TerrainTypes } from "../world/tile/tile-terrain";
// entities
import { City } from "../entities/city/city";
// manager
import { AbstractManager } from "./abstract-manager";
// gui
import { PlayerInfoDialog } from "../gui/dialogs/player-dialog";
import { CityDialog } from "../gui/dialogs/city-dialog";
// babylon
import * as BABYLON from 'babylonjs';
import { WorldTools } from "../../helpers/world-tools";
import { IUnitArmy, UnitArmy } from "../entities/unit/army/army";
import { SingleArmyDialog } from "../gui/dialogs/army-dialog";

export class SelectionManager extends AbstractManager {
    source: WorldSelection = new WorldSelection(null);
    _clickedPos: WorldPos;
    _cursorPos: WorldPos = new WorldPos(20, 20);

    constructor() {
        super("Selection");
    }

    setSourceToPlayer() {
        if (globalThis.PLAYER) {
            this.source.setToPlayer();
            const playerPosStr = globalThis.PLAYER.pos.asString();
            this.log(`source set to player at ${playerPosStr}`);
            return;
        }
        this.warn(`cant set source to player: Player not found`);
    }

    moveCursorTo(worldPos: WorldPos) {
        this._cursorPos = worldPos;
        const tile = globalThis.MAP.tileAtPos(worldPos);
        const tilePosVec3 = WorldTools.worldToScenePos(tile.pos);
        // update cursor pos
        if (tile && this.cursorMesh) {
            const offsetY = TerrainTypes.heightOf(tile.terrainChar) /2;
            this.cursorMesh.position.set(tilePosVec3.x, offsetY, tilePosVec3.z);
        }
    }

    movePlayerToCursor() {
        globalThis.PLAYER_M.movePlayerTo(this._cursorPos)
    }

    /* ON ENTITY CLICK */
    onCityClicked(city: City) {
        this.log(`click on city ${city.name}`);
        // set as selection source
        this.source.setToCity(city);
        this.onPosClick(city.pos);
        // show city dialog
        const cityDiag = new CityDialog(city);
        globalThis.GUI.showDialog(cityDiag);
    }

    onPlayerClicked() {
        this.onPosClick(globalThis.PLAYER.pos);
        globalThis.PLAYER_M.onPlayerClicked();
    }

    onTileClicked(tile: WorldTile) {
        this.log(`clicked on tile ${tile.terrainChar} at ${tile.pos.asString()}`);
        this.onPosClick(tile.pos);
        this.moveCursorTo(tile.pos);
    }

    onCursorClicked() {
        const cursorPosStr = this._cursorPos.asString();
        this.log(`clicked on cursor at ${cursorPosStr}: initialize movement check`);
        // check for possible movement
        this.movePlayerToCursor();
    }

    onPosClick(clickedPos: WorldPos) {
        this._clickedPos = clickedPos;
    }

    onPlayerMeshPlaced() {
        this.setSourceToPlayer();
    }

    onArmyClicked(army: UnitArmy) {
        this.log(`click on army \"${army.name}\" of ${army._ownerKey}`);
        // show army dialog
        const armyDiag = new SingleArmyDialog(army);
        globalThis.GUI.showDialog(armyDiag);
    }

    /* EVENT HANDLERS FROM OTHER MANAGERS */
    onMeshClicked(mesh: BABYLON.AbstractMesh) {
        // CITY click
        if (mesh.id.startsWith("city")) {
            const clickedCity = SelectionManager.GetCityByMesh(mesh);
            this.onCityClicked(clickedCity);
        // ARMY click
        } else if (mesh.id.startsWith("army")) {
            const clickedArmy: UnitArmy = SelectionManager.GetArmyByMesh(mesh);
            this.onArmyClicked(clickedArmy);
        // PLAYER caravan click
        } else if ( mesh.id.startsWith("player")) {
            this.onPlayerClicked();
        // TILE click on map
        } else if ( mesh.id.startsWith("tile")) {
            const tile = SelectionManager.GetTileByMesh(mesh);
            if (tile) { this.onTileClicked(tile); }
        // CURSOR click
        } else if( mesh.id.startsWith("cursor") ) {
            this.onCursorClicked();
        }
    }

    /* SELECTION MANAGER GETTERS */
    get cursorMesh(): BABYLON.AbstractMesh {
        return globalThis.WORLD.cursorMesh;
    }

    /* STATIC METHODS */
    static GetTileByMesh(tileMesh: BABYLON.AbstractMesh): WorldTile {
        let tilePosStr: string = tileMesh.id.split("-")[2];
        return globalThis.MAP.tileAtPos( WorldPos.FromString(tilePosStr, true) );
    }

    static GetCityByMesh(cityMesh: BABYLON.AbstractMesh): City {
        const cityKey= cityMesh.name.split("-")[1];
        return globalThis.CITIES.getByKey(cityKey);
    }

    static GetArmyByMesh(armyMesh: BABYLON.AbstractMesh): UnitArmy {
        const armyNum = parseInt(armyMesh.name.split("-")[1]);
        return globalThis.ARMIES.ByNum(armyNum);
    }

    onEvent(ev: IBaseEvent): void {
        super.onEvent(ev);
    }
}