// events
import { IBaseEvent, BaseEventListener } from '../../common/events';
import * as EV from "../game-events";
// trade
import { StoredResource } from '../entities/trade/resource/resource-storage';
import { TradeNode } from '../entities/trade/trade-node';
import { MarketTrade } from '../entities/trade/market-trade';
// world
import { ConnectedLocation, IConnectedLocation, ITradeLocation } from '../world/location';
// entities
import { TradeConnection } from '../entities/trade/connection/connection';
import { Road, IRoad } from '../world/road';
import { TradeResource } from '../entities/trade/resource/resource';
import { TimerState } from 'babylonjs/Legacy/legacy';
// manager
import { AbstractManager } from "./abstract-manager";

/* == TRADE MANAGER == */
export class TradeManager extends AbstractManager {
    worldTradeHistory: Array<MarketTrade> = [];
    // roads
    roads: Array<Road> = [];
    lastRoadNum: number = -1;

    constructor() {
        super("Trade");
    }
    
    onCitiesLoaded(): void {
        this.createRoadNetwork();
    }

    overrideSpecialResources() {
        
    }

    /* called when city delivery event is fired */
    onDelivery(source: ITradeLocation, target: ITradeLocation, storedRes: StoredResource) {
        const turnNum = globalThis.TURNS.turnNum();
        const marketTrade = new MarketTrade(turnNum, storedRes.resId, storedRes.profit, target, source);
        // add to world trade history
        this.worldTradeHistory.push(marketTrade);
        // increase city connection counter
        const connection = globalThis.CONN.fromTo(source, target);
        connection.addToTradeHistory(marketTrade, true);
        // add to trade history
        source.tradeHistory.add(marketTrade, true);
        target.tradeHistory.add(marketTrade, false);
    }

    onPlayerTrade(isSell: boolean, city: ITradeLocation, storedRes: StoredResource, profit: number) {

    }

    get hasSpecialResLoaded(): boolean {
        return globalThis.SPECIAL_RES && globalThis.SPECIAL_RES.length > 0;
    }

    /* inform gui over changed player monex */
    onPlayerCoinsChange(): void {
        globalThis.GUI.updateTopBar();
    }

    getRes(resId: string): TradeResource {
        if (resId.startsWith("X")) {
            const xResNum = parseInt( resId.replace("X", "") );
            return globalThis.SPECIAL_RES[xResNum];
        } else {
            return globalThis.RESOURCE_TYPES.get(resId);
        }
    }

    getResources(): Array<TradeResource> {
        let allResTypes: Array<TradeResource> = [];
        for (const resId of RESOURCE_TYPES.getIds()) {
            allResTypes.push(this.getRes(resId));
        }
        return allResTypes;
    }

    getResByName(resName: string): TradeResource {
        for (const tradeRes of globalThis.RESOURCE_TYPES.getAll()) {
            if (tradeRes.name == resName) { return tradeRes; }
        }
        return;
    }

    /* returns the resource name string ("Grain") of a resource with given resId */
    getResName(resId: string): string {
        const resObj = this.getRes(resId);
        return resObj.name;
    }

    getMarkets(): TradeNode[] {
        let markets: TradeNode[] = [];
        // city markets
        const cities = globalThis.CITIES.getAll();
        if (cities.length >= 0) {
            this.error("cant get markets; cities not loaded yet");
        }
        // add market to list
        for (const city of cities) { markets.push(city.tradeNode); }
        return markets;
    }

    /* CITY CONNECTIONS */
    private getCityConnection(from: IConnectedLocation, to: IConnectedLocation): TradeConnection {
        const connection = from.getConnectionTo(to);
        if (! connection) {
            this.warn(`cant get city connection: ${from.name} to ${to.name}`);
            return;
        }
        return connection;
    }

    /* ROADS */
    createRoadNetwork(): number {
        let roadCounter: number = 0;
        // create road objects
        for (const sourceCity of globalThis.CITIES.cities) {
            let sourceLocation: ConnectedLocation = sourceCity as ConnectedLocation;
            for (const connection of sourceLocation.connections) {
                let targetLocation: ConnectedLocation = connection.target as ConnectedLocation;
                this.addRoad(sourceLocation, targetLocation);
                roadCounter += 1;
            }
        }
        this.log(`${roadCounter} roads created`);
        // find road paths
        for (const road of this.roads) {
            road.findRoadPath();
        }
        this.log("created road paths");
        return roadCounter;
    }

    nextRoadNum(): number {
        const nextRoadNum: number = this.lastRoadNum +1;
        this.lastRoadNum = nextRoadNum;
        return nextRoadNum;
    }

    getRoad(roadNum: number): Road {
        return this.roads[roadNum];
    }

    getRoadFromTo(fromCity: IConnectedLocation, toCity: IConnectedLocation): IRoad {
        let roadsFromTarget: Array<IRoad> = fromCity.roads;
        for (const startCityRoad of roadsFromTarget) {
            const isToTargetCity: boolean = startCityRoad.target.equals(toCity);
            if(isToTargetCity) { return startCityRoad; }
        }
        this.warn(`cant find road between ${fromCity.name} & ${toCity.name}`);
        return null;
    }

    getReturnRoad(roadObj: IRoad): IRoad {
        return this.getRoadFromTo(roadObj.target, roadObj.start);
    }

    addRoad(startCity: IConnectedLocation, targetCity:IConnectedLocation): Road {
        let newRoad: Road = new Road(startCity, targetCity, this.nextRoadNum());
        startCity.roads.push(newRoad);
        targetCity.roads.push(newRoad);
        this.roads.push(newRoad);
        return newRoad;
    }

    /* GETTER SHORTFORMS */
    GetRes(resId: string): TradeResource { return this.getRes(resId); }
    GetResName(resId: string): string { return this.getRes(resId).name; }
}