// common
import { BaseEventListener } from "../../common/events";
import { GameOptions } from "../game-options";
import { Loggable } from "../../common/log";
import { IDataLoader, IDataGenerator } from "../../common/loader";
import { RandomInt, RandomElement } from "../../helpers/random";
// resources
/* CHANGE USED MAP HERE */
import * as MAPDATA from "../../../res/worlds/world_120_1600";
// world
import { WorldPos } from "../world/pos";
import { WorldTile } from "../world/tile/tile";
import { LandscapeFeature } from "../world/feature";
// models
import { City } from "../entities/city/city";
import { Empire, IEmpire } from "../entities/faction/empire";
import { Properties } from "../../common/properties";
import { TradeResource } from "../entities/trade/resource/resource";
import { ColorPalette } from "../../helpers/colorpalette";
import { Family, IFamily } from "../entities/faction/family";

export class WorldTilesLoader extends BaseEventListener implements IDataLoader {
    constructor() {
        super("WorldTileLoader");
    }

    load(): WorldTile[][] {
        let tileRows: WorldTile[][] = [];
        let createdTilesCount: number = 0;
        // for row
        for (let z = 0; z < MAPDATA.TILE_ROWS.length; z++) {
            const terrRow = MAPDATA.TILE_ROWS[z];
            let tileRow: WorldTile[] = [];
            for (let x = 0; x < terrRow.length; x++) {
                // find terrain char in array
                const pos = new WorldPos(x, z);
                const terrainChar = terrRow[pos.x];
                // create WorldTile & increase counter
                const worldTile = this.makeWorldTile(terrainChar, pos);
                createdTilesCount += 1;
                tileRow.push(worldTile);
            } // for x end
            tileRows.push(tileRow);
        } // for z end
        this.log(`${createdTilesCount} WorldTile objects loaded`);
        return tileRows;
    }

    /* returns WorldTile object with given pos & terrain */
    makeWorldTile(terrainChar: string, pos: WorldPos): WorldTile {
        const tile: WorldTile = new WorldTile(pos, terrainChar);
        // check whether tile is overwritten
        const terrainOverride = this.getTerrainOverwriteAtPos(pos);
        if (terrainOverride) { tile.terrainChar = terrainOverride; }
        return tile;
    }

    getTerrainOverwriteAtPos(pos: WorldPos): string {
        for (const tileOverwriteStr of MAPDATA.TILE_OVERRIDE) {
            const tileOverrideData = tileOverwriteStr.split(" ");
            const tileOverridePos = WorldPos.FromString(tileOverwriteStr);
            if (tileOverridePos.equals(pos)) { 
                const tileOverwriteTerrain = tileOverrideData[2];
                return tileOverwriteTerrain;
            }
        }
        return null;
    }
}

/* == LOADER: LANDSCAPE FEATURE == */
export class WorldLandscapeLoader extends BaseEventListener implements IDataLoader {
    constructor() {
        super("LandscapeLoader");
    }

    load(): LandscapeFeature[] {
        let features: LandscapeFeature[] = [];
        for (const featureStr of MAPDATA.LANDSCAPE) {
            const feature: LandscapeFeature = LandscapeFeature.FromString(featureStr)
            features.push(feature);
        }
        return features;
    }
}

/* == GENERATOR: CITIES == */
export class CityGenerator extends BaseEventListener implements IDataGenerator {
    private cities: Array<City> = [];

    constructor() {
        super("CityGenerator");
    }

    generate(): Array<City> {
        const maxCities = globalThis.OPTIONS.get(GameOptions.MAP_MAXCITIES);
        let generatedCities: Array<City> = [];
        for (let cityNum = 0; cityNum < maxCities; cityNum++) {
            let newCity: City = this.generateSingleCity(cityNum);
            generatedCities.push(newCity);
        }
        return generatedCities;
    }

    private generateSingleCity(cityNum: number): City {
        const generatedCity: City = City.MakeRandom();
        generatedCity.num = cityNum;
        return generatedCity;
    }
}

/* == LOADER: CITIES == */
export class CityLoader extends Loggable {
    static load(): Array<City> {
        let cities: Array<City> = [];
        for (const cityValues of MAPDATA.CITIES) {
            const loadedCity: City = City.FromValueArray(cityValues);
            cities.push(loadedCity);
        }
        return cities;
    }
}

/* == LOADER: RND START POS == */
export class RandomStartPosLoader extends Loggable {
    static load(): WorldPos {
        let positions: string[] = MAPDATA.RND_START_POS;
        let rndPosStr = RandomElement.of(positions);
        return WorldPos.FromString(rndPosStr);
    }
}

/* == LOADER: MAP EMPIRES == */
export class WorldEmpireLoader extends BaseEventListener implements IDataLoader {
    constructor() {
        super('EmpireLoader');
    }

    load(): IEmpire[] {
        let empires: IEmpire[] = [];
        for (const empireJson of MAPDATA.CUSTOM_EMPIRES) {
            if(! empireJson["key"]) {
                this.warn(`cant load empire ${empireJson.toString()} has no key property}`); continue;
            }
            const empireKey = empireJson["key"];
            const empireName = empireJson["name"];
            const empire = new Empire(empireKey, empireName);
            // set empire flag data
            const flagStr = empireJson["flag"];
            empire.flagData = flagStr;
            empire.updateFlag();
            empires.push(empire);
        }
        return empires;
    }
}

/* == LOADER: CUSTOM FAMILIES == */
export class CustomFamiliesLoader extends BaseEventListener implements IDataLoader {
    constructor() {
        super("CustomFamilyLoader");
    }

    load(): Array<IFamily> {
        if (! Object.keys(MAPDATA).includes("CUSTOM_FAMILIES")) {
            this.warn(`cant load custom families; none in map file`); return;
        }
        let customFamilies: Array<IFamily> = [];
        for (const familyJson of MAPDATA["CUSTOM_FAMILIES"]) {
            const name = familyJson["name"];
            const family = new Family(name, null, false);
            // set home key but load city later
            family._homeCityKey = familyJson["home"];
            // set optional properties
            family.customLanguage = familyJson["lang"];
            family.customStartPos = WorldPos.FromString(familyJson["customStartPos"])
            customFamilies.push(family);
        }
        return customFamilies;
    }
}


/* LOADER: RESOURCES */
export class SpecialResLoader extends BaseEventListener implements IDataLoader {
    constructor() {
        super('SpecialResLoader');
    }

    load(): Array<TradeResource> {
        let specialResources: Array<TradeResource> = [];
        if (MAPDATA["SPECIAL_RESOURCES"]) {
            const specialResJson = MAPDATA["SPECIAL_RESOURCES"];
            for (const singleResJson of specialResJson) {
                const tradeRes = TradeResource.FromJson(singleResJson);
                specialResources.push(tradeRes);
            }
        }
        return specialResources;
    }
}

/* LOADER: MAP PROPERTIES == */
export class WorldPropertiesLoader extends BaseEventListener implements IDataLoader {
    static load(): Properties {
        let mapProperties = new Properties("map");
        // start & end year
        mapProperties.addInteger("startYear", MAPDATA.STARTING_YEAR);
        mapProperties.addInteger("endYear", MAPDATA.END_YEAR);
        // seasons per turn
        mapProperties.addInteger("seasonsPerYear", MAPDATA.SEASONS_PER_YEAR);
        return mapProperties;
    }

}