// common
import { DawnbringerPalette, GrayTone } from '../../common/colors';
import { GameOptions } from '../game-options';
import { Loggable } from '../../common/log';
import { Properties } from '../../common/properties';
// world
import { ITile, TileMaterials, WorldTile } from './tile/tile';
import { TerrainTypes } from './tile/tile-terrain';
import { WorldTilesLoader, WorldLandscapeLoader, RandomStartPosLoader, WorldPropertiesLoader, SpecialResLoader } from '../managers/world-loaders';
import { WorldPos, Size, IPos } from './pos';
import { ILandscapeFeature, LandscapeFeature } from './feature';
// babylon
import * as BABYLON from 'babylonjs';

/* == INTERFACES == */
export interface IBaseMap {
    name: string;
    size: Size;
    properties: Properties;
}

export interface IWorldMap extends IBaseMap {
    name: string;
    size: Size;
    // start & end year
    startYear: number;
    endYear: number;
    // tiles
    playerStartingPos: WorldPos;
    tiles: ITile[][];
    // landscape features
    features: Array<ILandscapeFeature>;
    // methods
    tileAt(x: number, z:number): ITile;
}

/* == WORLD MAP MODEL == */
export class WorldMap extends Loggable implements IWorldMap {
    // properties
    name: string;
    size: Size = new Size(100, 100);
    properties: Properties = null;
    // start & end year
    startYear: number;
    endYear: number;
    // tiles
    terrainIds: Array<string>;
    tiles: WorldTile[][] = [];
    playerStartingPos: WorldPos = null;
    // landscape
    features: Array<LandscapeFeature> = [];
    // materials
    mats: TileMaterials = null;

    constructor(name: string, size: Size) {
        super("WorldMap");
        this.name = name;
        this.size = size;
        this.terrainIds = TerrainTypes.LIST;
        this.generateOrLoadTiles();
    }

    private generateOrLoadTiles(): void {
        let generate: boolean = globalThis.OPTIONS.getBool(GameOptions.GENERATE_LANDSCAPE);
        this.log(`loading tiles: random? ${generate}`);
        if (generate) {
            this._generateTiles();
        } else {
            this._loadProperties();
            this._loadSpecialRes();
            this._loadTiles();
            this._loadFeatures();
            this._loadStartingPos();
        }
    }

    private _loadProperties() {
        this.properties = WorldPropertiesLoader.load();
    }

    private _loadStartingPos(): void {
        this.playerStartingPos = RandomStartPosLoader.load();
    }

    private _generateTiles(): void {
        this.log(`tilemap: creating matrix`);
        let generatedTilesCount: number = 0;
        for (let z = 0; z < this.size.z; z++) {
            let tileRow: Array<WorldTile> = [];
            for (let x = 0; x < this.size.x; x++) {
                let terrain: string = this.getRandomTerrain();
                let tilePos: WorldPos = new WorldPos(x, z);
                let tile: WorldTile = new WorldTile(tilePos, terrain);
                // add tile & increase counter
                tileRow.push(tile);
                generatedTilesCount += 1;
            }  
            this.tiles.push(tileRow);
        }
        this.log(`${generatedTilesCount} tiles generated`);
    }

    private _loadTiles(): void {
        this.tiles = new WorldTilesLoader().load();
        // reset map size
        this.size = new Size(this.tiles[0].length, this.tiles.length);
        this.log(`resized to ${this.size.x}x${this.size.z}`);
    }

    private _loadFeatures(): void {
        const features = new WorldLandscapeLoader().load();
        this.features = features;
    }

    private _loadSpecialRes(): void {
        const specialRes = new SpecialResLoader().load();
        globalThis.SPECIAL_RES = specialRes;
    }

    private getRandomTerrain(): string {
        const landscapeTerrains = TerrainTypes.LANDSCAPE;
        const rndInt = Math.floor(Math.random() * landscapeTerrains.length);
        return landscapeTerrains[rndInt];
    }

    tileAt(x: number, z: number): WorldTile {
        let row: Array<WorldTile> = this.tiles[z];
        if (! row) {
            this.warn(`cant find row ${z} in ${this.tiles.length} tile array rows`); return;
        }
        let tileAt: WorldTile = row[x];
        if (! tileAt) {
            console.error(`cant get tile at ${x} ${z}`); return;
        }
        return tileAt;
    }

    tileAtPos(pos: WorldPos): WorldTile {
        return this.tileAt(pos.x, pos.z);
    }

    // ???
    public createMaterials(): void {
        /* == DEFINE TERRAIN COLORS HERE == */
        this.mats = new TileMaterials(this);
        this.mats.add(TerrainTypes.WATER, 
            this.paletteColor(DawnbringerPalette.CORNFLOWER), new GrayTone(0.5).toBabylon(), 128);
        this.mats.add(TerrainTypes.GRASS, 
            this.paletteColor(DawnbringerPalette.CHRISTI), new GrayTone(0.05).toBabylon(), 128);
        this.mats.add(TerrainTypes.HILL, 
            this.paletteColor(DawnbringerPalette.ROPE), new GrayTone(0.05).toBabylon(), 128);
        this.mats.add(TerrainTypes.FOREST, 
        this.paletteColor(DawnbringerPalette.DELL), new GrayTone(0.05).toBabylon(), 128);
        this.mats.add(TerrainTypes.STONE, 
            this.paletteColor(DawnbringerPalette.SMOKEYASH), new GrayTone(0.01).toBabylon(), 128);
        this.mats.add(TerrainTypes.BARREN, 
            this.paletteColor(DawnbringerPalette.DIMGRAY), new GrayTone(0.01).toBabylon(), 128);
        this.mats.add(TerrainTypes.MOUNTAIN, 
            this.paletteColor(DawnbringerPalette.TOPAZ), new GrayTone(0.01).toBabylon(), 128);
        this.mats.add(TerrainTypes.PEAK, 
            this.paletteColor(DawnbringerPalette.LIGHTSTEEL), new GrayTone(0.01).toBabylon(), 128);
        this.mats.add(TerrainTypes.CAVE, 
            this.paletteColor(DawnbringerPalette.SMOKEYASH), new GrayTone(0.01).toBabylon(), 128);
        this.mats.add(TerrainTypes.EARTH, this.paletteColor(DawnbringerPalette.ROPE));
        this.mats.add(TerrainTypes.DESERT, this.paletteColor(DawnbringerPalette.PANCHO));
        this.mats.add(TerrainTypes.SAVANNAH, this.paletteColor(DawnbringerPalette.RAINFOREST));
        this.mats.add(TerrainTypes.CITY, this.paletteColor(DawnbringerPalette.PANCHO));
        this.mats.add(TerrainTypes.ARCTIC, this.paletteColor(DawnbringerPalette.LIGHTSTEEL));
        this.mats.add(TerrainTypes.BRIDGE, this.paletteColor(DawnbringerPalette.VIKING));
        this.mats.add(TerrainTypes.LAVA, this.paletteColor(DawnbringerPalette.TAHITIGOLD));
        this.log(`created ${this.mats.length} tile materials`);
    }

    private paletteColor(rgb: Array<number>): BABYLON.Color3 {
        return new BABYLON.Color3(rgb[0], rgb[1], rgb[2]);
    }

    public getMaterial(key: string): BABYLON.StandardMaterial {
        return this.mats.get(key);
    }

    public get DEFAULT_SIZE(): Size {
        return new Size(100, 100);
    }

    public toString(): string {
        return `<WorldMap "${this.name}>`;
    }

    public makePos(x: number, z: number): WorldPos {
        return new WorldPos(x, z);
    }
}

/* GLOBAL METHODS */
export function tileAt(x: number, z: number): WorldTile {
    return globalThis.MAP.tileAt(x, z);
}