// common
import { Loggable } from '../../../common/log';
// world
import { HexDirection, IPos, WorldPos } from '../pos';
import { WorldMap } from '../map';
import { Road } from '../road';
// entities
import { City } from '../../entities/city/city';
import { IConnection } from '../../entities/trade/connection/connection';
// babylon
import * as BABYLON from 'babylonjs';
import { TerrainTypes } from './tile-terrain';

/* == INTERFACES == */
export interface ITile {
    pos: IPos;
    terrainChar: string;
    offset: number;
    connections: IConnection[];
    // meshes
    meshBlueprint: BABYLON.Mesh;
    meshInstance: BABYLON.InstancedMesh;
}

/* == TILE MODEL == */
export class WorldTile extends Loggable implements ITile {
    pos: WorldPos;
    terrainChar: string;
    offset: number = 0;
    // entities
    connections: IConnection[] = [];
    roads: Array<Road> = [];
    city: City = null;
    // meshes
    meshBlueprint: BABYLON.Mesh = null;
    meshInstance: BABYLON.InstancedMesh = null;

    constructor(pos: WorldPos, terrainChar: string) {
        super("WorldTile");
        this.pos = pos;
        this.terrainChar = terrainChar;
    }

    toString(): string {
        const posStr = `${this.pos.x}|${this.pos.z}`;
        return `<WorldTile "${this.terrainChar}" pos=${posStr}>`;
    }

    /* GETTERS */
    static get HEXSIZE(): number { return 3; }

    get isOnLand(): boolean {
        return !(this.terrainChar === TerrainTypes.WATER);
    }

    get meshName(): string {
        return `tile-${this.terrainChar}-${this.pos.asString()}`;
    }

    get isOnRoad(): boolean { return this.roads.length > 0; }

    get neighborTerrains(): string[] {
        let neighbTerrains: string[] = [];
        const neighbDirs: string[] = HexDirection.ALL;
        for (const dirStr of neighbDirs) {
            const dirObj: HexDirection = new HexDirection(dirStr);
            const relPos = dirObj.relPos;
            const absPos = new WorldPos(this.pos.x + relPos.x, this.pos.z + relPos.z);
            const neighbTile = globalThis.MAP.tileAtPos(absPos);
            const neighbTileTerrain = neighbTile.terrainChar;
            neighbTerrains.push(neighbTileTerrain);
        }
        return neighbTerrains;
    }
}


/* TILE MATERIALS */
export class TileMaterials {
    private _map: WorldMap;
    private _keys: Array<string> = [];
    private _mats: Array<BABYLON.StandardMaterial> = [];

    constructor(map: WorldMap) {
        this._map = map;
    }

    add(name: string, diffuseColor: BABYLON.Color3, specularColor?: BABYLON.Color3, specularPower?: number) {
        let mat: BABYLON.StandardMaterial = new BABYLON.StandardMaterial(name, globalThis.SCENE);
        mat.diffuseColor = diffuseColor;
        if (specularColor) {
            mat.specularColor = specularColor;
            if (specularPower) {
                mat.specularPower = specularPower;
            }
        }
        this._keys.push(name);
        this._mats.push(mat);
        // console.log(`material ${name} created: ${mat}`);
    }

    get(key: string): BABYLON.StandardMaterial {
        const matIndex: number = this._keys.indexOf(key);
        const mat = this._mats[matIndex];
        return mat;
    }

    public get length(): number {
        return this._keys.length;
    }
}